import base64
import io
import os
from typing import TYPE_CHECKING, SupportsFloat

import numpy as np
from bec_lib import bec_logger
from ophyd import PositionerBase, Staged

from ophyd_devices.sim.sim_camera import SimCamera
from ophyd_devices.sim.sim_frameworks.assets.default_image import DEFAULT_IMAGE
from ophyd_devices.sim.sim_frameworks.device_proxy import DeviceProxy

try:
    from PIL import Image
except ImportError as e:
    raise Exception(
        "PIL/pillow is not available - please install ophyd_devices with dev dependencies to enable this simulation"
    ) from e
logger = bec_logger.logger


class StageCameraProxy(DeviceProxy):
    """This Proxy class scans an ROI over an image based on some positioners, as if
    a stage was being moved in front of some camera. The sample config expects positioners
    samx and samy to exist in the device manager."""

    def __init__(self, name, *args, device_manager=None, **kwargs):
        self._file_source = (
            ""  # absolute path to an image file to use, or by default use the file from ./assets
        )
        self._staged = Staged.no
        self._roi_fraction: float = 0.15
        self._image: Image.Image | None = None
        self._shape: tuple[int, int] | None = None
        self._name = name

        self._x_roi_fraction: float
        self._y_roi_fraction: float
        self._image_size: tuple[int, int]
        self._motors: tuple[PositionerBase, PositionerBase]

        super().__init__(name, *args, device_manager=device_manager, **kwargs)

    def _validate_motors_from_config(self):
        ref_motors: tuple[list[str], ...] = self.config[self._device_name]["ref_motors"]
        logger.debug(f"using reference_motors {ref_motors} for camera view simulation")
        ref_motor_0 = self.device_manager.devices.get(ref_motors[0]).obj
        ref_motor_1 = self.device_manager.devices.get(ref_motors[1]).obj
        if ref_motor_0 is None:
            raise ValueError(f"{self._name}: device {ref_motor_0} doesn't exist in device manager")
        elif ref_motor_1 is None:
            raise ValueError(f"{self._name}: device {ref_motor_1} doesn't exist in device manager")
        else:
            self._motors: tuple[PositionerBase, PositionerBase] = (ref_motor_0, ref_motor_1)

    def _update_device_config(self, config: dict) -> None:
        super()._update_device_config(config)
        if len(self.config.keys()) > 1:
            raise RuntimeError(
                f"The current implementation of device {self.name} can only replay data for a single device. The config has information about multiple devices {config.keys()}"
            )
        logger.debug(f"{self._name} received config: {self.config}")
        self._device_name = list(self.config.keys())[0]
        roi_fraction = self.config[self._device_name].get("roi_fraction")
        if roi_fraction is not None:
            logger.debug(f"Updating roi_fraction on {self._name} to {roi_fraction}")
            if not isinstance(roi_fraction, SupportsFloat):
                raise ValueError('"roi_fraction" must be a number!')
            self._roi_fraction = roi_fraction

        self._validate_motors_from_config()

    def stop(self) -> None:
        """Stop the device."""
        ...

    def stage(self) -> list[object]:
        """Stage the device; loads the image from file."""
        self._device: SimCamera = self.device_manager.devices.get(self._device_name).obj
        self._shape = self._device.image_shape.get()
        shape_aspect_ratio = self._shape[0] / self._shape[1]
        if self._staged != Staged.no:
            return [self]
        try:
            self._load_image()
        except Exception as exc:
            raise type(e)(
                f"{self._name}: Could not open image file {self._file_source}, relative to {os.getcwd()}"
            ) from exc
        w, h = self._image.size
        self._x_roi_fraction = self._roi_fraction
        self._y_roi_fraction = h / w * self._roi_fraction / shape_aspect_ratio
        self._staged = Staged.yes
        return [self]

    def unstage(self) -> list[object]:
        """Unstage the device"""
        self._image = None
        self._staged = Staged.no
        return [self]

    def _load_image(self):
        """Try loading the image from the filesystem"""
        try:
            if self._file_source == "":
                logger.debug(f"{self._name} is using the default image")
                self._image = Image.open(io.BytesIO((base64.b64decode(DEFAULT_IMAGE))))
            else:
                self._image = Image.open(self.file_source)
            self._image.load()
        except Exception as e:
            raise type(e)(
                f"Make sure you have set the image path in the device config for {self._name}: - currently it is '{self._file_source}'"
            ) from e

    def _compute(self, *args, **kwargs):
        """Compute the image.

        Returns:
            np.ndarray: Image.
        """
        logger.debug("{self._name}: compute called.")

        if self._staged == Staged.no:
            logger.debug("")
            return np.zeros((*self._shape, 3))
        elif self._image is None:
            raise ValueError(
                f"{self._name}: Something went wrong - expected an image to have been loaded"
            )

        def get_positioner_fraction_along_limits(positioner: PositionerBase):
            if (limits := positioner.limits) == [0, 0] or limits[0] == limits[1]:
                raise ValueError(
                    f"Device {positioner} must have limits set to be used as an axis for the camera view simulation"
                )
            return (positioner.position - limits[0]) / (limits[1] - limits[0])

        x, y = (get_positioner_fraction_along_limits(m) for m in self._motors)
        w, h = self._image.size

        # x increases rightwards from the image origin
        cropped_x_min_px = x * (1 - self._x_roi_fraction) * w
        cropped_x_max_px = (x * (1 - self._x_roi_fraction) + self._x_roi_fraction) * w
        # y increases downard from the image origin
        cropped_y_min_px = h - (y * (1 - self._y_roi_fraction) * h)
        cropped_y_max_px = h - ((y * (1 - self._y_roi_fraction) + self._y_roi_fraction) * h)

        cropped_image = self._image.crop(
            (cropped_x_min_px, cropped_y_max_px, cropped_x_max_px, cropped_y_min_px)
        )
        scaled_image = cropped_image.resize(self._shape)

        return np.array(scaled_image)



from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as Dcpt

from ophyd_devices.sim.sim_positioner import SimPositioner
from ophyd_devices.sim.sim_signals import SetableSignal as SynSignal


class SynDeviceSubOPAAS(Device):
    zsub = Cpt(SimPositioner, name="zsub")


class SynDeviceOPAAS(Device):
    x = Cpt(SimPositioner, name="x")
    y = Cpt(SimPositioner, name="y")
    z = Cpt(SynDeviceSubOPAAS, name="z")


class SynDynamicComponents(Device):
    messages = Dcpt({f"message{i}": (SynSignal, None, {"name": f"msg{i}"}) for i in range(1, 6)})



"""Module to test the ophyd device config templates interface."""

from ophyd_devices.interfaces.device_config_templates.ophyd_templates import (
    OPHYD_DEVICE_TEMPLATES,
    EpicsMotorDeviceConfigTemplate,
    EpicsSignalDeviceConfigTemplate,
    EpicsSignalRODeviceConfigTemplate,
    EpicsSignalWithRBVDeviceConfigTemplate,
)


def test_interfaces_OPHYD_DEVICE_TEMPLATES():
    """Test that the OPHYD_DEVICE_TEMPLATES dictionary is correctly defined."""
    # Group level checks
    assert "EpicsMotor" in OPHYD_DEVICE_TEMPLATES
    assert "EpicsSignal" in OPHYD_DEVICE_TEMPLATES
    assert "CustomDevice" in OPHYD_DEVICE_TEMPLATES

    # Component level checks
    assert "EpicsMotor" in OPHYD_DEVICE_TEMPLATES["EpicsMotor"]
    template_info = OPHYD_DEVICE_TEMPLATES["EpicsMotor"]["EpicsMotor"]
    assert "ophyd.EpicsMotor" == template_info.get("deviceClass", "")
    assert issubclass(template_info.get("deviceConfig", object), EpicsMotorDeviceConfigTemplate)

    assert "EpicsSignalRO" in OPHYD_DEVICE_TEMPLATES["EpicsSignal"]
    template_info = OPHYD_DEVICE_TEMPLATES["EpicsSignal"]["EpicsSignalRO"]
    assert "ophyd.EpicsSignalRO" == template_info.get("deviceClass", "")
    assert issubclass(template_info.get("deviceConfig", object), EpicsSignalRODeviceConfigTemplate)

    assert "EpicsSignalWithRBV" in OPHYD_DEVICE_TEMPLATES["EpicsSignal"]
    template_info = OPHYD_DEVICE_TEMPLATES["EpicsSignal"]["EpicsSignalWithRBV"]
    assert "ophyd.EpicsSignalWithRBV" == template_info.get("deviceClass", "")
    assert issubclass(
        template_info.get("deviceConfig", object), EpicsSignalWithRBVDeviceConfigTemplate
    )

    assert "EpicsSignal" in OPHYD_DEVICE_TEMPLATES["EpicsSignal"]
    template_info = OPHYD_DEVICE_TEMPLATES["EpicsSignal"]["EpicsSignal"]
    assert "ophyd.EpicsSignal" == template_info.get("deviceClass", "")
    assert issubclass(template_info.get("deviceConfig", object), EpicsSignalDeviceConfigTemplate)

    assert "CustomDevice" in OPHYD_DEVICE_TEMPLATES["CustomDevice"]
    template_info = OPHYD_DEVICE_TEMPLATES["CustomDevice"]["CustomDevice"]
    assert template_info.get("deviceConfig", None) == {}


def test_interfaces_templates():
    """Test the individual device config templates."""
    # pylint: disable=unsupported-assignment-operation
    # Epics Motor Template
    assert "prefix" in EpicsMotorDeviceConfigTemplate.model_fields
    assert "EPICS IOC prefix" in EpicsMotorDeviceConfigTemplate.model_fields["prefix"].description
    assert "limits" in EpicsMotorDeviceConfigTemplate.model_fields
    assert "limits" in EpicsMotorDeviceConfigTemplate.model_fields["limits"].description

    # Epics Signal Template
    assert "read_pv" in EpicsSignalDeviceConfigTemplate.model_fields
    assert "EPICS read PV" in EpicsSignalDeviceConfigTemplate.model_fields["read_pv"].description
    assert "write_pv" in EpicsSignalDeviceConfigTemplate.model_fields
    assert "EPICS write PV" in EpicsSignalDeviceConfigTemplate.model_fields["write_pv"].description

    # Epics Signal RO Template
    assert "read_pv" in EpicsSignalRODeviceConfigTemplate.model_fields
    assert "EPICS read PV" in EpicsSignalRODeviceConfigTemplate.model_fields["read_pv"].description

    # Epics Signal With RBV Template
    assert "prefix" in EpicsSignalWithRBVDeviceConfigTemplate.model_fields
    assert (
        "EPICS IOC prefix"
        in EpicsSignalWithRBVDeviceConfigTemplate.model_fields["prefix"].description
    )



"""This module tests the EpicsMotor and EpicsMotorEC classes for custom
PSI motor integration from the ophyd_devices.devices.psi_motor module."""

from __future__ import annotations

import threading
from unittest import mock

import ophyd
import pytest

from ophyd_devices.devices.psi_motor import EpicsMotor, EpicsMotorEC, SpmgStates
from ophyd_devices.tests.utils import MockPV, patch_dual_pvs


@pytest.fixture(scope="function")
def mock_epics_motor():
    """Fixture to create a mock EpicsMotor instance."""
    name = "test_motor"
    prefix = "SIM:MOTOR"
    with mock.patch.object(ophyd, "cl") as mock_cl:
        mock_cl.get_pv = MockPV
        mock_cl.thread_class = threading.Thread
        motor = EpicsMotor(name=name, prefix=prefix)
        motor.wait_for_connection()
        patch_dual_pvs(motor)
        yield motor


@pytest.fixture(scope="function")
def mock_epics_motor_ec():
    """Fixture to create a mock EpicsMotorEC instance."""
    name = "test_motor_ec"
    prefix = "SIM:MOTOR:EC"
    with mock.patch.object(ophyd, "cl") as mock_cl:
        mock_cl.get_pv = MockPV
        mock_cl.thread_class = threading.Thread
        motor_ec = EpicsMotorEC(name=name, prefix=prefix)
        motor_ec.wait_for_connection()
        patch_dual_pvs(motor_ec)
        yield motor_ec


def test_epics_motor_limits_raise(mock_epics_motor):
    """Test the move method of EpicsMotor."""
    motor = mock_epics_motor
    motor.user_setpoint._metadata["lower_ctrl_limit"] = -10
    motor.user_setpoint._metadata["upper_ctrl_limit"] = 10
    motor.low_limit_travel.put(-10)
    motor.high_limit_travel.put(10)
    with pytest.raises(ophyd.utils.LimitError):
        motor.move(-15)


def test_epics_motor_move(mock_epics_motor):
    """Test the move method of EpicsMotor."""
    motor = mock_epics_motor
    assert motor.user_readback.get() == 0
    assert motor.user_setpoint.get() == 0
    motor.user_setpoint._metadata["lower_ctrl_limit"] = -10
    motor.user_setpoint._metadata["upper_ctrl_limit"] = 10
    motor.motor_mode.put(SpmgStates.GO)
    # Now this should raise... in theory...
    motor.user_readback._read_pv.set_severity(ophyd.utils.epics_pvs.AlarmSeverity.NO_ALARM)
    motor.user_readback._metadata["severity"] = ophyd.utils.epics_pvs.AlarmSeverity.NO_ALARM
    motor.user_readback._read_pv.set_alarm_status(ophyd.utils.epics_pvs.AlarmStatus.NO_ALARM)
    motor.user_readback._metadata["status"] = ophyd.utils.epics_pvs.AlarmStatus.NO_ALARM
    # Set alarm of motor to
    status = motor.move(5)
    assert status.done is False
    assert motor.user_readback.get() == 0
    assert motor.user_setpoint.get() == 5
    motor.motor_done_move._read_pv.mock_data = 0
    for ii in range(1, 6):
        motor.user_readback._read_pv.mock_data = ii
        if ii == 5:
            motor.motor_done_move._read_pv.mock_data = 1
        else:
            motor.motor_done_move._read_pv.mock_data = 0

    status.wait(timeout=5)
    assert status.done is True
    assert motor.user_readback.get() == 5
    assert motor.user_setpoint.get() == 5

    # Now this should raise...
    motor.user_readback._read_pv.set_severity(ophyd.utils.epics_pvs.AlarmSeverity.MAJOR)
    motor.user_readback._metadata["severity"] = ophyd.utils.epics_pvs.AlarmSeverity.MAJOR
    motor.user_readback._read_pv.set_alarm_status(ophyd.utils.epics_pvs.AlarmStatus.READ)
    motor.user_readback._metadata["status"] = ophyd.utils.epics_pvs.AlarmStatus.READ
    status = motor.move(10)
    motor.motor_done_move._read_pv.mock_data = 0
    assert status.done is False
    assert motor.user_readback.get() == 5
    assert motor.user_setpoint.get() == 10
    motor.user_readback._read_pv.mock_data = 7
    motor.motor_done_move._read_pv.mock_data = 0
    assert status.done is False
    motor.user_readback._read_pv.mock_data = 10
    motor.motor_done_move._read_pv.mock_data = 1
    with pytest.raises(RuntimeError):
        status.wait(timeout=5)
        assert status.done is True
        assert isinstance(status.exception(), RuntimeError)


def test_move_epics_motor_ec(mock_epics_motor_ec):
    """Test the move method of EpicsMotorEC."""
    motor_ec = mock_epics_motor_ec
    motor_ec.user_setpoint._metadata["lower_ctrl_limit"] = -10
    motor_ec.user_setpoint._metadata["upper_ctrl_limit"] = 10
    motor_ec.user_readback._read_pv.set_severity(ophyd.utils.epics_pvs.AlarmSeverity.NO_ALARM)
    motor_ec.user_readback._metadata["severity"] = ophyd.utils.epics_pvs.AlarmSeverity.NO_ALARM
    motor_ec.user_readback._read_pv.set_alarm_status(ophyd.utils.epics_pvs.AlarmStatus.NO_ALARM)
    motor_ec.user_readback._metadata["status"] = ophyd.utils.epics_pvs.AlarmStatus.NO_ALARM

    assert motor_ec.user_readback.get() == 0
    assert motor_ec.user_setpoint.get() == 0

    motor_ec.error._read_pv.mock_data = 0
    with mock.patch.object(motor_ec.log, "warning") as mock_log_warning:
        # Should not log any warning
        status = motor_ec.move(5)
        motor_ec.motor_done_move._read_pv.mock_data = 0
        assert mock_log_warning.call_count == 0
        motor_ec.user_readback._read_pv.mock_data = 5
        motor_ec.motor_done_move._read_pv.mock_data = 1
        status.wait(timeout=5)
        assert status.done is True

        # If error exists, this should raise a warning
        motor_ec.error._read_pv.mock_data = 1
        status = motor_ec.move(8)
        motor_ec.motor_done_move._read_pv.mock_data = 0
        assert status.done is False
        assert mock_log_warning.call_count == 1
        motor_ec.user_readback._read_pv.mock_data = 8
        motor_ec.motor_done_move._read_pv.mock_data = 1
        with pytest.raises(RuntimeError):
            status.wait(timeout=5)
            assert status.done is True
            assert isinstance(status.exception(), RuntimeError)

        motor_ec.error._read_pv.mock_data = 0
        # Note: Position has to be > current position from 8 to 9...
        motor_ec.high_interlock._read_pv.mock_data = 1
        status = motor_ec.move(9)
        motor_ec.motor_done_move._read_pv.mock_data = 0
        assert status.done is False
        assert mock_log_warning.call_count == 2
        motor_ec.user_readback._read_pv.mock_data = 2
        motor_ec.motor_done_move._read_pv.mock_data = 1
        motor_ec.high_interlock._read_pv.mock_data = 0
        status.wait(timeout=5)
        assert status.done is True

        # Attempting to move below active LLS should raise a warning
        motor_ec.low_interlock._read_pv.mock_data = 1
        status = motor_ec.move(-2)
        motor_ec.motor_done_move._read_pv.mock_data = 0
        assert status.done is False
        assert mock_log_warning.call_count == 3
        motor_ec.user_readback._read_pv.mock_data = -2
        motor_ec.motor_done_move._read_pv.mock_data = 1
        status.wait(timeout=5)
        assert status.done is True
        motor_ec.low_interlock._read_pv.mock_data = 0


def test_epics_motor_high_limit_switch_raises(mock_epics_motor):
    """Test that moving beyond high limit switch raises an error."""
    motor = mock_epics_motor
    motor.user_setpoint._metadata["lower_ctrl_limit"] = -10
    motor.user_setpoint._metadata["upper_ctrl_limit"] = 10
    motor.high_limit_switch._read_pv.mock_data = 1  # Simulate high limit switch active
    with pytest.raises(ophyd.utils.LimitError):
        motor.move(15)



import threading
from unittest import mock

import ophyd
import pytest

from ophyd_devices.devices.undulator import UndulatorGap
from ophyd_devices.tests.utils import MockPV, patch_dual_pvs


@pytest.fixture(scope="function")
def mock_undulator():
    name = "undulator"
    prefix = "TEST:UNDULATOR"
    with mock.patch.object(ophyd, "cl") as mock_cl:
        mock_cl.get_pv = MockPV
        mock_cl.thread_class = threading.Thread
        dev = UndulatorGap(name=name, prefix=prefix)
        patch_dual_pvs(dev)
        yield dev


@pytest.mark.parametrize(
    ["start", "end", "in_deadband_expected"],
    [
        (1.0, 1.0, True),
        (0, 1.0, False),
        (-0.004, 0.004, False),
        (-0.0027, -0.0023, True),
        (1, 1.0009, False),
        (1, 1.0007, True),
    ],
)
@mock.patch("ophyd_devices.devices.undulator.PVPositioner.move")
@mock.patch("ophyd_devices.devices.undulator.MoveStatus")
def test_instant_completion_within_deadband(
    mock_movestatus, mock_super_move, mock_undulator, start, end, in_deadband_expected
):
    mock_undulator.select_control._read_pv.mock_data = 1
    mock_undulator._position = start
    mock_undulator.move(end)

    if in_deadband_expected:
        mock_movestatus.assert_called_with(mock.ANY, mock.ANY, done=True, success=True)
    else:
        mock_movestatus.assert_not_called()
        mock_super_move.assert_called_once()


def test_undulator_raises_when_disabled(mock_undulator):
    mock_undulator.select_control._read_pv.mock_data = 0
    with pytest.raises(PermissionError) as e:
        mock_undulator.move(5)
    assert e.match("Undulator is operator controlled!")


def test_undulator_stop_call(mock_undulator):
    mock_undulator.select_control._read_pv.mock_data = 1
    mock_undulator.stop_signal.put(0)
    mock_undulator.stop()
    assert mock_undulator.stop_signal.get() == 1
    mock_undulator.stop_signal.put(0)
    mock_undulator.select_control._read_pv.mock_data = 0
    # Error should just be logged, not raised.
    mock_undulator.stop()
    assert mock_undulator.stop_signal.get() == 0



"""Utility handler to run tasks (function, conditions) in an asynchronous fashion."""

import ctypes
import operator
import threading
import traceback
import uuid
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Literal

from bec_lib.file_utils import get_full_path
from bec_lib.logger import bec_logger
from bec_lib.utils.import_utils import lazy_import_from
from ophyd import Device, Signal
from ophyd.status import AndStatus, DeviceStatus, MoveStatus, Status, StatusBase, SubscriptionStatus

if TYPE_CHECKING:  # pragma: no cover
    from bec_lib.messages import ScanStatusMessage
else:
    # TODO: put back normal import when Pydantic gets faster
    ScanStatusMessage = lazy_import_from("bec_lib.messages", ("ScanStatusMessage",))


__all__ = [
    "CompareStatus",
    "TransitionStatus",
    "AndStatus",
    "DeviceStatus",
    "MoveStatus",
    "Status",
    "StatusBase",
    "SubscriptionStatus",
]

logger = bec_logger.logger

set_async_exc = ctypes.pythonapi.PyThreadState_SetAsyncExc

OP_MAP = {
    "==": operator.eq,
    "!=": operator.ne,
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
}


class CompareStatus(SubscriptionStatus):
    """
    Status class to compare a signal value against a given value.
    The comparison is done using the specified operation, which can be one of
    '==', '!=', '<', '<=', '>', '>='. If the value is a string, only '==' and '!=' are allowed.
    The status is finished when the comparison is true.

    Args:
        signal: The device signal to compare.
        value: The value to compare against.
        operation: The operation to use for comparison. Defaults to '=='.
        event_type: The type of event to trigger on comparison. Defaults to None (default sub).
        timeout: The timeout for the status. Defaults to None (indefinite).
        settle_time: The time to wait for the signal to settle before comparison. Defaults to 0.
        run: Whether to run the status callback on creation or not. Defaults to True.
    """

    def __init__(
        self,
        signal: Signal,
        value: float | int | str,
        *,
        operation: Literal["==", "!=", "<", "<=", ">", ">="] = "==",
        event_type=None,
        timeout: float = None,
        settle_time: float = 0,
        run: bool = True,
    ):
        if isinstance(value, str):
            if operation not in ("==", "!="):
                raise ValueError(
                    f"Invalid operation: {operation} for string comparison. Must be '==' or '!='."
                )
        if operation not in ("==", "!=", "<", "<=", ">", ">="):
            raise ValueError(
                f"Invalid operation: {operation}. Must be one of '==', '!=', '<', '<=', '>', '>='."
            )
        self._signal = signal
        self._value = value
        self._operation = operation
        super().__init__(
            device=signal,
            callback=self._compare_callback,
            timeout=timeout,
            settle_time=settle_time,
            event_type=event_type,
            run=run,
        )

    def _compare_callback(self, value, **kwargs) -> bool:
        """Callback for subscription status"""
        return OP_MAP[self._operation](value, self._value)


class TransitionStatus(SubscriptionStatus):
    """
    Status class to monitor transitions of a signal value through a list of specified transitions.
    The status is finished when all transitions have been observed in order. The keyword argument
    `strict` determines whether the transitions must occur in strict order or not.
    If `raise_states` is provided, the status will raise an exception if the signal value matches
    any of the values in `raise_states`.

    Args:
        signal: The device signal to monitor.
        transitions: A list of values to transition through.
        strict: Whether the transitions must occur in strict order. Defaults to True.
        raise_states: A list of values that will raise an exception if encountered. Defaults to None.
        run: Whether to run the status callback on creation or not. Defaults to True.
        event_type: The type of event to trigger on transition. Defaults to None (default sub).
        timeout: The timeout for the status. Defaults to None (indefinite).
        settle_time: The time to wait for the signal to settle before comparison. Defaults to 0.
    """

    def __init__(
        self,
        signal: Signal,
        transitions: list[float | int | str],
        *,
        strict: bool = True,
        raise_states: list[float | int | str] | None = None,
        run: bool = True,
        event_type=None,
        timeout: float = None,
        settle_time: float = 0,
    ):
        self._signal = signal
        if not isinstance(transitions, list):
            raise ValueError(f"Transitions must be a list of values. Received: {transitions}")
        self._transitions = transitions
        self._index = 0
        self._strict = strict
        self._raise_states = raise_states if raise_states else []
        super().__init__(
            device=signal,
            callback=self._compare_callback,
            timeout=timeout,
            settle_time=settle_time,
            event_type=event_type,
            run=run,
        )

    def _compare_callback(self, old_value, value, **kwargs) -> bool:
        """Callback for subscription Status"""
        if value in self._raise_states:
            self.set_exception(
                ValueError(
                    f"Transition raised an exception: {value}. "
                    f"Expected transitions: {self._transitions}."
                )
            )
            return False
        if self._index == 0:
            if value == self._transitions[0]:
                self._index += 1
        else:
            if self._strict:
                if (
                    old_value == self._transitions[self._index - 1]
                    and value == self._transitions[self._index]
                ):
                    self._index += 1
            else:
                if value == self._transitions[self._index]:
                    self._index += 1
        return self._is_finished()

    def _is_finished(self) -> bool:
        """Check if the status is finished"""
        return self._index >= len(self._transitions)


class TaskState(str, Enum):
    """Possible task states"""

    NOT_STARTED = "not_started"
    RUNNING = "running"
    TIMEOUT = "timeout"
    ERROR = "error"
    COMPLETED = "completed"
    KILLED = "killed"


class TaskKilledError(Exception):
    """Exception raised when a task thread is killed"""


class TaskStatus(DeviceStatus):
    """Thin wrapper around StatusBase to add information about tasks"""

    def __init__(self, device: Device, *, timeout=None, settle_time=0, done=None, success=None):
        super().__init__(
            device=device, timeout=timeout, settle_time=settle_time, done=done, success=success
        )
        self._state = TaskState.NOT_STARTED
        self._task_id = str(uuid.uuid4())

    @property
    def state(self) -> str:
        """Get the state of the task"""
        return self._state.value

    @state.setter
    def state(self, value: TaskState):
        self._state = TaskState(value)

    @property
    def task_id(self) -> str:
        """Get the task ID"""
        return self._task_id


class TaskHandler:
    """Handler to manage asynchronous tasks"""

    def __init__(self, parent: Device):
        """Initialize the handler"""
        self._tasks = {}
        self._parent = parent
        self._lock = threading.RLock()

    def submit_task(
        self,
        task: Callable,
        task_args: tuple | None = None,
        task_kwargs: dict | None = None,
        run: bool = True,
    ) -> TaskStatus:
        """Submit a task to the task handler.

        Args:
            task: The task to run.
            run: Whether to run the task immediately.
        """
        task_args = task_args if task_args else ()
        task_kwargs = task_kwargs if task_kwargs else {}
        task_status = TaskStatus(device=self._parent)
        thread = threading.Thread(
            target=self._wrap_task,
            args=(task, task_args, task_kwargs, task_status),
            name=f"task {task_status.task_id}",
            daemon=True,
        )
        self._tasks.update({task_status.task_id: (task_status, thread)})
        if run is True:
            self.start_task(task_status)
        return task_status

    def start_task(self, task_status: TaskStatus) -> None:
        """Start a task,

        Args:
            task_status: The task status object.
        """
        thread = self._tasks[task_status.task_id][1]
        if thread.is_alive():
            logger.warning(f"Task with ID {task_status.task_id} is already running.")
            return
        task_status.state = TaskState.RUNNING
        thread.start()

    def _wrap_task(
        self, task: Callable, task_args: tuple, task_kwargs: dict, task_status: TaskStatus
    ):
        """Wrap the task in a function"""
        try:
            task(*task_args, **task_kwargs)
        except TimeoutError as exc:
            content = traceback.format_exc()
            logger.warning(
                (
                    f"Timeout Exception in task handler for task {task_status.task_id},"
                    f" Traceback: {content}"
                )
            )
            task_status.state = TaskState.TIMEOUT
            task_status.set_exception(exc)
        except TaskKilledError as exc:
            exc = exc.__class__(
                f"Task {task_status.task_id} was killed. ThreadID:"
                f" {self._tasks[task_status.task_id][1].ident}"
            )
            content = traceback.format_exc()
            logger.warning(
                (
                    f"TaskKilled Exception in task handler for task {task_status.task_id},"
                    f" Traceback: {content}"
                )
            )
            task_status.state = TaskState.KILLED
            task_status.set_exception(exc)
        except Exception as exc:  # pylint: disable=broad-except
            content = traceback.format_exc()
            logger.warning(
                f"Exception in task handler for task {task_status.task_id}, Traceback: {content}"
            )
            task_status.state = TaskState.ERROR
            task_status.set_exception(exc)
        else:
            task_status.state = TaskState.COMPLETED
            task_status.set_finished()
        finally:
            with self._lock:
                self._tasks.pop(task_status.task_id, None)

    def kill_task(self, task_status: TaskStatus) -> None:
        """Kill the thread

        task_status: The task status object.
        """
        thread = self._tasks[task_status.task_id][1]
        exception_cls = TaskKilledError

        ident = ctypes.c_long(thread.ident)
        exc = ctypes.py_object(exception_cls)
        try:
            res = set_async_exc(ident, exc)
            if res == 0:
                raise ValueError("Invalid thread ID")
            if res > 1:
                set_async_exc(ident, None)
                logger.warning(f"Exception raise while kille Thread {ident}; return value: {res}")
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f"Exception raised while killing thread {ident}: {e}")

    def shutdown(self):
        """Shutdown all tasks of task handler"""
        with self._lock:
            for info in self._tasks.values():
                self.kill_task(info[0])


class FileHandler:
    """Utility class for file operations."""

    def get_full_path(
        self, scan_status_msg: ScanStatusMessage, name: str, create_dir: bool = True
    ) -> str:
        """Get the file path.

        Args:
            scan_info_msg: The scan info message.
            name: The name of the file.
            create_dir: Whether to create the directory.
        """
        return get_full_path(scan_status_msg=scan_status_msg, name=name, create_dir=create_dir)



"""
This module contains the base class for custom device integrations at PSI.
Please check the device section in BEC's developer documentation
(https://bec.readthedocs.io/en/latest/) for more information about device integration.
"""

import os
import threading
import time
import traceback
from typing import Generic, TypeVar

from bec_lib.file_utils import FileWriter
from bec_lib.logger import bec_logger
from ophyd import Device, DeviceStatus, Kind
from ophyd.device import Staged

from ophyd_devices.utils import bec_utils
from ophyd_devices.utils.bec_scaninfo_mixin import BecScaninfoMixin
from ophyd_devices.utils.errors import DeviceStopError, DeviceTimeoutError

logger = bec_logger.logger

T = TypeVar("T", bound="BECDeviceBase")


class BECDeviceBaseError(Exception):
    """Error class for BECDeviceBase."""


class CustomPrepare(Generic[T]):
    """Custom prepare class for beamline specific logic.

    This class provides a set of hooks for beamline specific logic to be implemented.
    BECDeviceBase will be injected as the parent device.

    To implement custom logic, inherit from this class and implement the desired methods.
    If the __init__ method is overwritten, please ensure the proper initialisation.
    It is important to pass the parent device to the custom prepare class as this
    will allow the custom prepare class to access the parent device with all its signals.
    """

    def __init__(self, *_args, parent: T = None, **_kwargs) -> None:
        """
        Initialize the custom prepare class.

        Args:
            parent (BECDeviceBase): The parent device which gives access to all methods and signals.
        """
        self.parent = parent

    def on_init(self) -> None:
        """
        Hook for beamline specific logic during class initialization.

        This method is called during the initialization of the device class.
        It should not be used to set any of the class's signals as they may not
        be connected yet.
        """
        pass

    def on_wait_for_connection(self) -> None:
        """
        Hook for beamline specific logic during the wait_for_connection method.

        This method is called after Ophyd's wait_for_connection method was called,
        meaning that signals will be connected at this point. It can be used to check
        signal values, or set default values for those.
        """

    def on_stage(self) -> None:
        """
        Hook for beamline specific logic during the stage method.

        This method is called during the stage method of the device class.
        It is used to implement logic in preparation for a scan."""

    def on_unstage(self) -> None:
        """
        Hook for beamline specific logic during the unstage method.

        This method is called during the unstage method of the device class.
        It is used to implement logic to clean up the device after a scan.
        """

    def on_stop(self) -> None:
        """Hook for beamline specific logic during the stop method."""

    def on_trigger(self) -> None | DeviceStatus:
        """
        Hook for beamline specific logic during the trigger method.

        This method has to be non-blocking, and if time-consuming actions are necessary,
        they should be implemented asynchronously. Please check the wait_with_status
        method to implement asynchronous checks.

        The softwareTrigger config value needs to be set to True to indicate to BEC
        that the device should be triggered from the software during the scan.

        Returns:
            DeviceStatus: DeviceStatus object that BEC will use to check if the trigger was successful
        """

    def on_pre_scan(self) -> None:
        """
        Hook for beamline specific logic during the pre_scan method.

        This method is called from BEC just before the scan core starts, and be used
        to execute time-critical actions, e.g. arming a detector in case there is a risk of timing out.
        Note, this method should not be used to implement blocking logic.
        """

    def on_complete(self) -> None | DeviceStatus:
        """
        Hook for beamline specific logic during the complete method.

        This method is used to check whether the device has successfully completed its acquisition.
        For example, a detector may want to check if it has received the correct number of frames, or
        if it's data backend has finished writing the data to disk. The method should be implemented
        asynchronously. Please check the wait_with_status method on how to implement asynchronous checks.

        Returns:
            DeviceStatus: DeviceStatus object that BEC will use to check if the device has successfully completed
        """

    def on_kickoff(self) -> None | DeviceStatus:
        """
        Hook for beamline specific logic during the kickoff method.

        This method is called to kickoff the flyer acquisition. BEC will not call this method in general
        for its scans, but only if the scan explicitly implements this. The method should be non-blocking,
        and if time-consuming actions are necessary, they should be implemented asynchronously.
        Please check the wait_with_status method on how to implement asynchronous checks.

        Returns:
            DeviceStatus: DeviceStatus object that BEC will use to check if the kickoff was successful
        """

    def wait_for_signals(
        self,
        signal_conditions: list[tuple],
        timeout: float,
        check_stopped: bool = False,
        interval: float = 0.05,
        all_signals: bool = False,
    ) -> bool:
        """
        Utility method to implement waiting for signals to reach a certain condition. It accepts
        a list of conditions passed as tuples of executable calls for conditions (get_current_state, condition) to check.
        It can further be specified if all signals should be True or if any signal should be True.
        If the timeout is reached, it will return False.

        Args:
            signal_conditions (list[tuple]): tuple of executable calls for conditions (get_current_state, condition) to check
            timeout (float): timeout in seconds
            check_stopped (bool): True if stopped flag should be checked
            interval (float): interval in seconds
            all_signals (bool): True if all signals should be True, False if any signal should be True

        Returns:
            bool: True if all signals are in the desired state, False if timeout is reached

        Example:
            >>> self.wait_for_signals(signal_conditions=[(self.acquiring.get, False)], timeout=5, interval=0.05, check_stopped=True, all_signals=True)
        """

        timer = 0
        while True:
            checks = [
                get_current_state() == condition
                for get_current_state, condition in signal_conditions
            ]
            if check_stopped is True and self.parent.stopped is True:
                return False
            if (all_signals and all(checks)) or (not all_signals and any(checks)):
                return True
            if timer > timeout:
                return False
            time.sleep(interval)
            timer += interval

    def wait_with_status(
        self,
        signal_conditions: list[tuple],
        timeout: float,
        check_stopped: bool = False,
        interval: float = 0.05,
        all_signals: bool = False,
        exception_on_timeout: Exception = None,
    ) -> DeviceStatus:
        """
        Utility method to implement asynchronous waiting for signals to reach a certain condition.
        It accepts a list of conditions passed as tuples of executable calls.

        Please check the wait_for_signals method as it is used to implement the waiting logic.
        It returns a DeviceStatus object that can be used to check if the asynchronous action is done
        through 'status.done', and if it was successful through 'status.success'. An exception can be
        passed to the method which will be raised if the timeout is reached. If the device was stopped
        during the waiting, a DeviceStopError will be raised.

        Args:
            signal_conditions (list[tuple]): tuple of executable calls for conditions (get_current_state, condition) to check
            timeout (float): timeout in seconds
            check_stopped (bool): True if stopped flag should be checked
            interval (float): interval in seconds
            all_signals (bool): True if all signals should be True, False if any signal should be True
            exception_on_timeout (Exception): Exception to raise on timeout

        Returns:
            DeviceStatus: DeviceStatus object to check the state of the asynchronous action (status.done, status.success)

        Example:
            >>> status = self.wait_with_status(signal_conditions=[(self.acquiring.get, False)], timeout=5, interval=0.05, check_stopped=True, all_signals=True)
        """
        if exception_on_timeout is None:
            exception_on_timeout = DeviceTimeoutError(
                f"Timeout error for {self.parent.name} while waiting for signals {signal_conditions}"
            )

        status = DeviceStatus(self.parent)

        # utility function to wrap the wait_for_signals function
        def wait_for_signals_wrapper(
            status: DeviceStatus,
            signal_conditions: list[tuple],
            timeout: float,
            check_stopped: bool,
            interval: float,
            all_signals: bool,
            exception_on_timeout: Exception,
        ):
            try:
                result = self.wait_for_signals(
                    signal_conditions, timeout, check_stopped, interval, all_signals
                )
                if result:
                    status.set_finished()
                else:
                    if self.parent.stopped:
                        # INFO This will execute a callback to the parent device.stop() method
                        status.set_exception(exc=DeviceStopError(f"{self.parent.name} was stopped"))
                    else:
                        # INFO This will execute a callback to the parent device.stop() method
                        status.set_exception(exc=exception_on_timeout)
            # pylint: disable=broad-except
            except Exception as exc:
                content = traceback.format_exc()
                logger.warning(
                    f"Error in wait_for_signals in {self.parent.name}; Traceback: {content}"
                )
                # INFO This will execute a callback to the parent device.stop() method
                status.set_exception(exc=exc)

        thread = threading.Thread(
            target=wait_for_signals_wrapper,
            args=(
                status,
                signal_conditions,
                timeout,
                check_stopped,
                interval,
                all_signals,
                exception_on_timeout,
            ),
            daemon=True,
        )
        thread.start()
        return status


class BECDeviceBase(Device):
    """
    Base class for custom device integrations at PSI. This class wraps around the ophyd's standard
    set of methods, providing hooks for custom logic to be implemented in the custom_prepare_cls.

    Please check the device section in BEC's developer documentation
    (https://bec.readthedocs.io/en/latest/) for more information about device integration.
    """

    custom_prepare_cls = CustomPrepare

    # All possible subscription types that the Device Manager subscribes to
    # Run the command _run_subs(sub_type=self.SUB_VALUE, value=value) to trigger
    # the subscription of type value. Please be aware that the signature of
    # the subscription callbacks needs to be matched.
    SUB_READBACK = "readback"
    SUB_VALUE = "value"
    SUB_DONE_MOVING = "done_moving"
    SUB_MOTOR_IS_MOVING = "motor_is_moving"
    SUB_PROGRESS = "progress"
    SUB_FILE_EVENT = "file_event"
    SUB_DEVICE_MONITOR_1D = "device_monitor_1d"
    SUB_DEVICE_MONITOR_2D = "device_monitor_2d"
    _default_sub = SUB_VALUE

    def __init__(
        self,
        name: str,
        prefix: str = "",
        kind: Kind | None = None,
        parent=None,
        device_manager=None,
        **kwargs,
    ):
        """
        Initialize the device.

        Args:
            name (str): name of the device
            prefix (str): prefix of the device
            kind (Kind): kind of the device
            parent (Device): parent device
            device_manager (DeviceManager): device manager. BEC will inject the device manager as a dependency.
        """
        super().__init__(prefix=prefix, name=name, kind=kind, parent=parent, **kwargs)
        self._stopped = False
        self.service_cfg = None
        self.scaninfo = None
        self.filewriter = None

        if not issubclass(self.custom_prepare_cls, CustomPrepare):
            raise BECDeviceBaseError(
                f"Custom prepare class must be subclass of CustomDetectorMixin, provided: {self.custom_prepare_cls}"
            )
        self.custom_prepare = self.custom_prepare_cls(parent=self, **kwargs)

        if device_manager:
            self._update_service_config()
            self.device_manager = device_manager
        else:
            # If device manager is not provided through dependency injection
            # A mock device manager is created. This is necessary to be able
            # to use the device class in a standalone context without BEC.
            self.device_manager = bec_utils.DMMock()
            base_path = kwargs["basepath"] if "basepath" in kwargs else "."
            self.service_cfg = {"base_path": os.path.abspath(base_path)}

        self.connector = self.device_manager.connector
        self._update_scaninfo()
        self._update_filewriter()
        self._init()

    @property
    def stopped(self) -> bool:
        """Property to indicate if the device was stopped."""
        return self._stopped

    @stopped.setter
    def stopped(self, value: bool) -> None:
        self._stopped = value

    def _update_filewriter(self) -> None:
        """Initialise the file writer utility class."""
        self.filewriter = FileWriter(service_config=self.service_cfg, connector=self.connector)

    def _update_scaninfo(self) -> None:
        """Initialise the utility class to get scan metadata from BEC."""
        self.scaninfo = BecScaninfoMixin(self.device_manager)
        self.scaninfo.load_scan_metadata()

    def _update_service_config(self) -> None:
        """Update the Service Config. This has the info on where REDIS and other services are running."""
        # pylint: disable=import-outside-toplevel
        from bec_lib.bec_service import SERVICE_CONFIG

        if SERVICE_CONFIG:
            self.service_cfg = SERVICE_CONFIG.model.file_writer.model_dump()
            return
        self.service_cfg = {"base_path": os.path.abspath(".")}

    def check_scan_id(self) -> None:
        """
        Check if the scan ID has changed, if yes, set the stopped property to True.

        The idea is that if the scan ID on the device and the scan ID from BEC are different,
        the device is out of sync with the scan and should be stopped.
        """
        old_scan_id = self.scaninfo.scan_id
        self.scaninfo.load_scan_metadata()
        if self.scaninfo.scan_id != old_scan_id:
            self.stopped = True

    def _init(self) -> None:
        """
        Hook for beamline specific logic during class initialization.

        Please to not set any of the class's signals during intialisation, but
        instead use the wait_for_connection.
        """
        self.custom_prepare.on_init()

    def wait_for_connection(self, all_signals=False, timeout=5) -> None:
        """
        Thin wrapper around ophyd's wait_for_connection.

        Calling Ophyd's wait_for_connection will ensure that signals are connected. BEC
        will call this method if a device is not connected yet. This method should
        be used to set up default values for signals, or to check if the device signals
        are in the expected state.

        Args:
            all_signals (bool): True if all signals should be considered. Default is False.
            timeout (float): timeout in seconds. Default is 5 seconds.
        """
        super().wait_for_connection(all_signals, timeout)
        self.custom_prepare.on_wait_for_connection()

    def stage(self) -> list[object]:
        """
        Thin wrapper around ophyd's stage, the method called in preparation for a scan.

        Stage is idempotent, if staged twice it should raise (we let ophyd.Device handle the raise here).
        Other that that, we reset the stopped property in case the device was stopped before, and
        pull the latest scan metadata from BEC. Ater that, we allow for beamline specific logic to
        be implemented through the custom_prepare.on_stage method.

        Returns:
            list(object): list of objects that were staged
        """
        if self._staged != Staged.no:
            return super().stage()
        self.stopped = False
        self.scaninfo.load_scan_metadata()
        self.custom_prepare.on_stage()
        return super().stage()

    def unstage(self) -> list[object]:
        """
        This wrapper around ophyd's unstage method, which is called to clean up the device.

        It must be possible to call unstage multiple times without raising an exception. It should
        be implemented to clean up the device if it is in a staged state.

        Beamline specific logic can be implemented through the custom_prepare.on_unstage method.

        Returns:
            list(object): list of objects that were unstaged
        """
        self.check_scan_id()
        self.custom_prepare.on_unstage()
        self.stopped = False
        return super().unstage()

    def pre_scan(self) -> None:
        """
        Pre-scan is a method introduced by BEC that is not native to the ophyd interface.

        This method is called from BEC just before the scan core starts, and therefore should only
        implement time-critical actions. I.e. Arming a detector in case there is a risk of timing out.
        """
        self.custom_prepare.on_pre_scan()

    def trigger(self) -> DeviceStatus:
        """
        Thin wrapper around the trigger method of the device, for which the config value
        softwareTrigger needs to be set to True, which will indicate to BEC that the device
        should be triggered from the software during the scan.

        Custom logic should be implemented non-blocking, i.e. be fast, or implemented asynchroniously.

        Returns:
            DeviceStatus: DeviceStatus object that BEC will use to check if the trigger was successful.
        """
        # pylint: disable=assignment-from-no-return
        status = self.custom_prepare.on_trigger()
        if isinstance(status, DeviceStatus):
            return status
        return super().trigger()

    def complete(self) -> DeviceStatus:
        """
        Thin wrapper around ophyd's complete method. Complete is called once the scan core
        has finished, but before the scan is closed. It will be called before unstage.
        It can also be used for fly scans to track the status of the flyer, and indicate if the
        flyer has completed.

        The method is used to check whether the device has successfully completed the acquisition.
        Actions are implemented in custom_prepare.on_complete since they are beamline specific.

        This method has to be non-blocking. If checks are necessary, they should be implemented asynchronously.

        Returns:
            DeviceStatus: DeviceStatus object that BEC will use to check if the device has successfully completed.
        """
        # pylint: disable=assignment-from-no-return
        status = self.custom_prepare.on_complete()
        if isinstance(status, DeviceStatus):
            return status
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def stop(self, *, success=False) -> None:
        """Stop the device.

        Args:
            success (bool): Argument from ophyd's stop method. Default is False.
        """
        self.custom_prepare.on_stop()
        super().stop(success=success)
        self.stopped = True

    def kickoff(self) -> DeviceStatus:
        """
        This wrapper around Ophyd's kickoff method.

        Kickoff is a method native to the flyer interface of Ophyd. It is called to
        start the flyer acquisition. This method is not called by BEC in general, but
        only if the scan explicitly implements this.

        The method should be non-blocking, and if time-consuming actions are necessary,
        they should be implemented asynchronously.

        Returns:
            DeviceStatus: DeviceStatus object that BEC will use to check if the kickoff was successful.
        """
        # pylint: disable=assignment-from-no-return
        status = self.custom_prepare.on_kickoff()
        if isinstance(status, DeviceStatus):
            return status
        status = DeviceStatus(self)
        status.set_finished()
        return status



"""
Module for undulator control
"""

from __future__ import annotations

import enum

from bec_lib.logger import bec_logger
from ophyd import EpicsSignal, EpicsSignalRO, PVPositioner
from ophyd.device import Component as Cpt
from ophyd.signal import DEFAULT_CONNECTION_TIMEOUT, DEFAULT_WRITE_TIMEOUT
from ophyd.status import MoveStatus

logger = bec_logger.logger


class UNDULATORCONTROL(int, enum.Enum):
    """
    Enum for undulator control modes.
    """

    OPERATOR = 0
    BEAMLINE = 1


class UndulatorSetointSignal(EpicsSignal):
    """
    SLS Undulator setpoint control
    """

    def put(
        self,
        value,
        force=False,
        connection_timeout=DEFAULT_CONNECTION_TIMEOUT,
        callback=None,
        use_complete=None,
        timeout=DEFAULT_WRITE_TIMEOUT,
        **kwargs,
    ):
        """
        Put a value to the setpoint PV.

        If the undulator is operator controlled, it will not move.
        """
        if self.parent.select_control.get() == UNDULATORCONTROL.OPERATOR.value:
            raise PermissionError("Undulator is operator controlled!")
        return super().put(
            value,
            force=force,
            connection_timeout=connection_timeout,
            callback=callback,
            use_complete=use_complete,
            timeout=timeout,
            **kwargs,
        )


class UndulatorStopSignal(EpicsSignal):
    """
    SLS Undulator stop signal"""

    def put(
        self,
        value,
        force=False,
        connection_timeout=DEFAULT_CONNECTION_TIMEOUT,
        callback=None,
        use_complete=None,
        timeout=DEFAULT_WRITE_TIMEOUT,
        **kwargs,
    ):
        """
        Put a value to the setpoint PV.

        If the undulator is operator controlled, it will not move.
        """
        if self.parent.select_control.get() == UNDULATORCONTROL.OPERATOR.value:
            return None
        return super().put(
            value,
            force=force,
            connection_timeout=connection_timeout,
            callback=callback,
            use_complete=use_complete,
            timeout=timeout,
            **kwargs,
        )


class UndulatorGap(PVPositioner):
    """
    SLS Undulator gap control
    """

    setpoint = Cpt(UndulatorSetointSignal, suffix="GAP-SP")
    readback = Cpt(EpicsSignalRO, suffix="GAP-RBV", kind="hinted", auto_monitor=True)

    stop_signal = Cpt(UndulatorStopSignal, suffix="STOP")
    done = Cpt(EpicsSignalRO, suffix="DONE", auto_monitor=True)

    select_control = Cpt(EpicsSignalRO, suffix="SCTRL", auto_monitor=True)

    def __init__(
        self,
        prefix="",
        *,
        limits=None,
        name=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        egu="",
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            limits=limits,
            name=name,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            egu=egu,
            **kwargs,
        )
        # Make the default alias for the user_readback the name of the
        # motor itself.
        self.readback.name = self.name
        self.readback._metadata["write_access"] = False

    def move(self, position, wait=True, timeout=None, moved_cb=None):

        # If it is operator controlled, undulator will not move.
        if self.select_control.get() == 0:
            raise PermissionError("Undulator is operator controlled!")

        # If it is already there, undulator will not move. The done flag
        # will not change, the moving change callback will not be called.
        # The status will not change.
        if self._position is not None and abs(position - self._position) < 0.0008:
            logger.info(
                f"Undulator gap {self.name} already close to position {position}, not moving."
            )
            status = MoveStatus(self, position, done=True, success=True)
            return status

        return super().move(position, wait=wait, timeout=timeout, moved_cb=moved_cb)



import os
import sys
from unittest import mock

import bec_lib

from ophyd_devices.utils.static_device_test import StaticDeviceTest, TestResult, launch


def test_static_device_test():
    config_path = os.path.join(os.path.dirname(bec_lib.__file__), "configs", "demo_config.yaml")
    sys.argv = ["", "--config", config_path, "--connect"]
    launch()


def test_static_device_test_with_config_dict():
    """First device is okay, second one is not"""
    device_dict = {
        "waveform": {
            "readoutPriority": "async",
            "deviceClass": "ophyd_devices.SimWaveform",
            "deviceConfig": {
                "waveform_shape": 1000,
                "sim_init": {
                    "model": "GaussianModel",
                    "params": {"amplitude": 100, "center": 500, "sigma": 50},
                },
            },
            "deviceTags": ["detector"],
            "enabled": True,
            "readOnly": False,
            "softwareTrigger": True,
        },
        "wrong": {"this is not corect": 0},
    }
    test = StaticDeviceTest(config_dict=device_dict)
    ret = test.run_with_list_output(connect=False)
    assert len(ret) == 2
    assert ret[0].name == "waveform"
    assert ret[0].success is True
    assert ret[0].message == "waveform is OK"
    assert ret[1].name == "wrong"
    assert ret[1].success is False
    assert isinstance(ret[1].message, str)


def test_static_device_test_TestResults():
    result = TestResult(
        name="test_device", success=True, message="Device is OK", config_is_valid=True
    )
    assert result.name == "test_device"
    assert result.success is True
    assert result.message == "Device is OK"
    assert result.config_is_valid is True



import math
import os
import time
from pathlib import Path

import h5py
import hdf5plugin
import numpy as np


class H5Writer:
    """Utility class to write data from device to disk"""

    def __init__(self, file_path: str = None, h5_entry: str = None):
        self.file_path = file_path
        self.h5_entry = h5_entry
        self.h5_file = None
        self.file_handle = None
        self.data_container = []

    def create_dir(self):
        """Create directory if it does not exist"""
        file_path = str(Path(self.file_path).resolve())
        base_dir = os.path.dirname(file_path)
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)

    def receive_data(self, data: any):
        """Store data to be written to h5 file"""
        self.data_container.append(data)
        if len(self.data_container) > 2:
            self.write_data()

    def on_stage(self, file_path: str, h5_entry: str):
        """Prepare to write data to h5 file"""
        self.data_container.clear()
        self.file_path = file_path
        self.h5_entry = h5_entry
        self.create_dir()
        # Create file and truncate if it exists
        with h5py.File(self.file_path, "w") as f:
            pass

    def on_complete(self):
        """Write data to h5 file"""
        if len(self.data_container) > 0:
            self.write_data()

    def on_unstage(self):
        """Close file handle"""

    def write_data(self):
        """Write data to h5 file. If the scan is started, the file will be truncated first"""
        with h5py.File(self.file_path, "a") as f:
            dataset = self.h5_entry
            value = self.data_container
            if isinstance(value, list):
                shape = (
                    value[0].shape if hasattr(value[0], "shape") else (len(value), len(value[0]))
                )
                shape = (None, *shape)
            if dataset not in f:
                f.create_dataset(
                    dataset, data=np.array(value), maxshape=shape, chunks=True, **hdf5plugin.LZ4()
                )
            else:
                f[dataset].resize((f[dataset].shape[0] + len(value)), axis=0)
                f[dataset][-len(value) :] = np.array(value)
            self.data_container.clear()


class LinearTrajectory:
    def __init__(
        self, initial_position, final_position, max_velocity, acceleration, initial_time=None
    ):
        self.initial_position = initial_position
        self.final_position = final_position
        self.max_velocity = abs(max_velocity)
        self.acceleration = abs(acceleration)
        self.initial_time = initial_time if initial_time is not None else time.time()
        self._velocity_profile = []
        self.ended = False

        self.distance = self.final_position - self.initial_position
        self.direction = np.sign(self.distance)
        self.distance = abs(self.distance)

        # Calculate time to reach max velocity and the distance covered during this time
        self.time_to_max_velocity = self.max_velocity / self.acceleration
        self.distance_to_max_velocity = 0.5 * self.acceleration * self.time_to_max_velocity**2

        if self.distance < 2 * self.distance_to_max_velocity:
            # If the distance is too short to reach max velocity (triangular profile)
            self.time_accel = np.sqrt(self.distance / self.acceleration)
            self.time_decel = self.time_accel
            self.time_const_vel = 0
            self.total_time = 2 * self.time_accel
        else:
            # If the distance allows reaching max velocity (trapezoidal profile)
            self.time_accel = self.time_to_max_velocity
            self.time_decel = self.time_to_max_velocity
            self.time_const_vel = (
                self.distance - 2 * self.distance_to_max_velocity
            ) / self.max_velocity
            self.total_time = self.time_accel + self.time_const_vel + self.time_decel

    def _get_velocity_at_time(self, dt):
        if dt <= self.time_accel:
            # Acceleration phase
            return self.direction * self.acceleration * dt
        elif dt <= self.time_accel + self.time_const_vel:
            # Constant velocity phase
            return self.direction * self.max_velocity
        elif dt <= self.total_time:
            # Deceleration phase
            return self.direction * self.acceleration * (self.total_time - dt)
        else:
            return 0

    def _get_pos_at_time(self, dt):
        if dt <= self.time_accel:
            # Acceleration phase
            return self.initial_position + self.direction * 0.5 * self.acceleration * dt**2
        elif dt <= self.time_accel + self.time_const_vel:
            # Constant velocity phase
            return self.initial_position + self.direction * (
                self.distance_to_max_velocity + self.max_velocity * (dt - self.time_accel)
            )
        elif dt <= self.total_time:
            # Deceleration phase
            return (
                self.final_position
                - self.direction * 0.5 * self.acceleration * (self.total_time - dt) ** 2
            )
        else:
            return self.final_position

    def position(self, t=None):
        if t is None:
            t = time.time()
        dt = t - self.initial_time

        if dt < 0:
            raise ValueError("Time cannot be before initial time.")

        current_position = self._get_pos_at_time(dt)
        current_velocity = self._get_velocity_at_time(dt)

        self._velocity_profile.append([t, current_velocity])

        if dt > self.total_time:
            self.ended = True

        return current_position

    @property
    def velocity_profile(self):
        return np.array(self._velocity_profile)

    def plot_trajectory(self):
        # visual check of LinearTrajectory class
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError("plot_trajectory requires matplotlib to be installed")

        initial_time = time.time()
        trajectory = LinearTrajectory(
            self.initial_position,
            self.final_position,
            self.max_velocity,
            self.acceleration,
            initial_time,
        )

        # Simulate some time points
        positions = []
        times = []
        while not trajectory.ended:
            times.append(time.time())
            pos = trajectory.position(times[-1])
            positions.append(pos)
            time.sleep(0.01)

        # Plotting
        plt.figure(figsize=(12, 6))

        # Plot velocity profile
        plt.subplot(1, 2, 1)
        plt.plot(
            trajectory.velocity_profile[:, 0] - initial_time,
            trajectory.velocity_profile[:, 1],
            label="Velocity",
        )
        plt.xlabel("Time (s)")
        plt.ylabel("Velocity (m/s)")
        plt.title("Velocity Profile")
        plt.grid(True)
        plt.legend()

        # Plot position profile
        plt.subplot(1, 2, 2)
        plt.plot(np.array(times) - initial_time, positions, label="Position")
        plt.xlabel("Time (s)")
        plt.ylabel("Position (m)")
        plt.title("Position Profile")
        plt.grid(True)
        plt.legend()

        plt.tight_layout()
        plt.show()


def stop_trajectory(trajectory, stop_time=None):
    """Return a trajectory that starts to decelerate at stop_time,
    with same characteristics as given trajectory"""
    if stop_time is None:
        stop_time = time.time()
    # Calculate current position and velocity at the stop time
    dt = stop_time - trajectory.initial_time
    current_position = trajectory._get_pos_at_time(dt)
    current_velocity = trajectory._get_velocity_at_time(dt)

    dec_time = abs(current_velocity / trajectory.acceleration)
    decel_distance = trajectory.direction * (
        current_velocity * current_velocity / (2 * trajectory.acceleration)
    )

    # Create a new trajectory from current position-decel_distance to a stop at current position+decel_distance
    new_trajectory = LinearTrajectory(
        current_position - decel_distance,
        current_position + decel_distance,
        abs(current_velocity),
        trajectory.acceleration,
        stop_time - dec_time,
    )
    # Keep velocity profile data, so it is possible to check the move
    new_trajectory._velocity_profile = trajectory._velocity_profile[:]

    return new_trajectory



from .sim_camera import SimCamera
from .sim_flyer import SimFlyer

SynFlyer = SimFlyer
from .sim_frameworks import SlitProxy
from .sim_monitor import SimMonitor
from .sim_positioner import SimPositioner
from .sim_signals import ReadOnlySignal, SetableSignal
from .sim_test_devices import SimPositionerWithCommFailure, SimPositionerWithController
from .sim_waveform import SimWaveform
from .sim_xtreme import SynXtremeOtf



"""
Test module for DXP integration, i.e. Falcon, XMAP and Mercury detectors.
This also includes EpicsMCARecord for data recording of multichannel analyzers.
"""

import threading
from unittest import mock

import ophyd
import pytest
from ophyd import Component as Cpt

from ophyd_devices.devices.dxp import (
    ADBase,
    EpicsDXPFalcon,
    EpicsDxpFalconMapping,
    EpicsDXPFalconMultiElementSystem,
    EpicsDXPMultiElementSystem,
    EpicsMCARecord,
    Falcon,
    Mercury,
    xMAP,
)
from ophyd_devices.tests.utils import MockPV, patch_dual_pvs

# from ophyd.mca import EpicsDXPMapping


# pylint:disable=redefined-outer-name
# pylint: disable=protected-access


class TestFalcon(Falcon):
    """Test class for Falcon device integration together with mca record and dxp faclon."""

    mca1 = Cpt(EpicsMCARecord, "mca1")
    dxp1 = Cpt(EpicsDXPFalcon, "dxp1")


@pytest.fixture(scope="function")
def mock_falcon():
    """Fixture to create a mock Falcon device for testing."""
    name = "mca"
    prefix = "test_falcon"
    with mock.patch.object(ophyd, "cl") as mock_cl:
        mock_cl.get_pv = MockPV
        mock_cl.thread_class = threading.Thread
        ddg = TestFalcon(name=name, prefix=prefix)
        patch_dual_pvs(ddg)
        yield ddg


@pytest.fixture(scope="function")
def mock_xmap():
    """Fixture to create a mock xMAP device for testing."""
    name = "mca"
    prefix = "test_xmap"
    with mock.patch.object(ophyd, "cl") as mock_cl:
        mock_cl.get_pv = MockPV
        mock_cl.thread_class = threading.Thread
        ddg = xMAP(name=name, prefix=prefix)
        patch_dual_pvs(ddg)
        yield ddg


@pytest.fixture(scope="function")
def mock_mercury():
    """Fixture to create a mock Mercury device for testing."""
    name = "mca"
    prefix = "test_mercury"
    with mock.patch.object(ophyd, "cl") as mock_cl:
        mock_cl.get_pv = MockPV
        mock_cl.thread_class = threading.Thread
        ddg = Mercury(name=name, prefix=prefix)
        patch_dual_pvs(ddg)
        yield ddg


def test_falcon(mock_falcon):
    """Test the Falcon device."""
    # Test the default values
    mock_falcon: TestFalcon
    assert mock_falcon.name == "mca"
    assert mock_falcon.prefix == "test_falcon"
    assert isinstance(mock_falcon, EpicsDXPFalconMultiElementSystem)
    assert isinstance(mock_falcon, EpicsDxpFalconMapping)
    assert isinstance(mock_falcon, ADBase)

    assert mock_falcon.hints == {"fields": []}

    assert mock_falcon.mca1.rois.roi0.read_attrs == ["count", "net_count"]
    assert mock_falcon.mca1.rois.roi0.configuration_attrs == [
        "label",
        "preset_count",
        "is_preset",
        "bkgnd_chans",
        "hi_chan",
        "lo_chan",
    ]


def test_falcon_trigger(mock_falcon):
    """Test the Falcon device trigger method."""
    mock_falcon: TestFalcon
    mock_falcon.erase_start.put(0)
    assert mock_falcon.erase_start.get() == 0
    status = mock_falcon.trigger()
    # Tagged trigger value, should be set to 1 when trigger is called
    assert mock_falcon.erase_start.get() == 1
    assert status.success is True
    assert status.done is True


def test_xmap(mock_xmap):
    """Test the xMAP device."""
    # Test the default values
    mock_xmap: xMAP
    assert mock_xmap.name == "mca"
    assert mock_xmap.prefix == "test_xmap"
    assert isinstance(mock_xmap, EpicsDXPMultiElementSystem)
    assert isinstance(mock_xmap, ADBase)


def test_mercury(mock_mercury):
    """Test the Mercury device."""
    # Test the default values
    mock_mercury: Mercury
    assert mock_mercury.name == "mca"
    assert mock_mercury.prefix == "test_mercury"
    assert isinstance(mock_mercury, EpicsDXPMultiElementSystem)
    # assert isinstance(mock_mercury, EpicsDXPMapping) # Not sure why this fails
    assert isinstance(mock_mercury, ADBase)


def test_xmap_trigger(mock_xmap):
    """Test the xMAP device trigger method."""
    mock_xmap: xMAP
    mock_xmap.erase_start.put(0)
    assert mock_xmap.erase_start.get() == 0
    status = mock_xmap.trigger()
    # Tagged trigger value, should be set to 1 when trigger is called
    assert mock_xmap.erase_start.get() == 1
    assert status.success is True
    assert status.done is True



from .devices.sls_devices import SLSInfo, SLSOperatorMessages
from .sim.sim_camera import SimCamera
from .sim.sim_monitor import SimMonitor, SimMonitorAsync

SynAxisMonitor = SimMonitor
SynGaussBEC = SimMonitor
from .sim.sim_positioner import SimLinearTrajectoryPositioner, SimPositioner

SynAxisOPAAS = SimPositioner
from .sim.sim_flyer import SimFlyer

SynFlyer = SimFlyer
from .sim.sim import SynDeviceOPAAS, SynDynamicComponents
from .sim.sim_frameworks import DeviceProxy, H5ImageReplayProxy, SlitProxy
from .sim.sim_signals import ReadOnlySignal
from .sim.sim_waveform import SimWaveform

SynSignalRO = ReadOnlySignal
from .devices.psi_motor import EpicsMotor, EpicsMotorEC
from .devices.softpositioner import SoftPositioner
from .interfaces.base_classes.psi_device_base import PSIDeviceBase
from .utils.bec_device_base import BECDeviceBase
from .utils.bec_signals import *
from .utils.dynamic_pseudo import ComputedSignal
from .utils.psi_device_base_utils import *
from .utils.static_device_test import launch



from abc import ABC, abstractmethod

from bec_lib import bec_logger
from ophyd import Component as Cpt
from ophyd import EpicsMotor
from typeguard import typechecked

from ophyd_devices.interfaces.protocols.bec_protocols import BECRotationProtocol
from ophyd_devices.utils.bec_utils import ConfigSignal

logger = bec_logger.logger


class OphtyRotationBaseError(Exception):
    """Exception specific for implmenetation of rotation stages."""


class OphydRotationBase(BECRotationProtocol, ABC):

    allow_mod360 = Cpt(ConfigSignal, name="allow_mod360", value=False, kind="config")

    def __init__(self, *args, **kwargs):
        """
        Base class to implement functionality specific for rotation devices.

        Childrens should override the instance attributes:
        - has_mod360
        - has_freerun
        - valid_rotation_modes

        """
        # pylint: disable=protected-access
        self._has_mod360 = False
        self._has_freerun = False
        self._valid_rotation_modes = []
        if "allow_mod360" in kwargs:
            if not isinstance(kwargs["allow_mod360"], bool):
                raise ValueError("allow_mod360 must be a boolean")
            self.allow_mod360.put(kwargs["allow_mod360"])
        super().__init__(*args, **kwargs)

    @abstractmethod
    def apply_mod360(self) -> None:
        """Method to apply the modulus 360 operation on the specific device.

        Childrens should override this method
        """

    @property
    def has_mod360(self) -> bool:
        """Property to check if the device has mod360 operation.

        ReadOnly property, childrens should override this method.
        """
        return self._has_mod360

    @property
    def has_freerun(self) -> bool:
        """Property to check if the device has freerun operation.

        ReadOnly property, childrens should override this method.
        """
        return self._has_freerun

    @property
    def valid_rotation_modes(self) -> list:
        """Method to get the valid rotation modes for the specific device."""
        return self._valid_rotation_modes

    @typechecked
    @valid_rotation_modes.setter
    def valid_rotation_modes(self, value: list[str]):
        """Method to set the valid rotation modes for the specific device."""
        self._valid_rotation_modes = value
        return self._valid_rotation_modes


# pylint: disable=too-many-ancestors
class EpicsRotationBase(OphydRotationBase, EpicsMotor):
    """Class for Epics rotation devices."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._has_freerun = True
        self._has_freerun = True
        self._valid_rotation_modes = ["target", "radiography"]

    def apply_mod360(self) -> None:
        """Apply modulos 360 operation for EpicsMotorRecord.

        EpicsMotor has the function "set_current_position" which can be used for this purpose.
        In addition, there is a check if mod360 is allowed and available.
        """
        if self.has_mod360 and self.allow_mod360.get():
            cur_val = self.user_readback.get()
            new_val = cur_val % 360
            try:
                self.set_current_position(new_val)
            except Exception as exc:
                error_msg = f"Failed to set new position {new_val} from {cur_val} on device {self.name} with error {exc}"
                raise OphtyRotationBaseError(error_msg) from exc
            return
        logger.info(
            f"Did not apply mod360 for device {self.name} with has_mod={self.has_mod360} and allow_mod={self.allow_mod360.get()}"
        )



from unittest import mock

import pytest
from bec_server.device_server.tests.utils import DMMock

from ophyd_devices.utils.dynamic_pseudo import ComputedSignal


@pytest.fixture
def device_manager_with_devices():
    dm = DMMock()
    dm.add_device("a")
    dm.add_device("b")
    device_mock = mock.MagicMock()
    device_mock.obj.readback.get.return_value = 20
    dm.devices["a"] = device_mock
    dm.devices["b"] = device_mock

    return dm


def test_computed_signal(device_manager_with_devices):
    signal = ComputedSignal(name="test", device_manager=device_manager_with_devices)
    assert signal.get() is None

    signal.compute_method = "def test(a, b): return a.get() + b.get()"
    signal.input_signals = ["a_readback", "b_readback"]
    assert signal.get() == 40

    # pylint: disable=protected-access
    assert callable(signal._compute_method)
    assert signal._compute_method_str == "def user_compute_method(a, b): return a.get() + b.get()"



from ophyd.quadem import QuadEM
from ophyd.sim import SynAxis, SynPeriodicSignal, SynSignal

from .epics_motor_ex import EpicsMotorEx
from .psi_motor import EpicsMotor, EpicsMotorEC
from .sls_devices import SLSInfo, SLSOperatorMessages
from .SpmBase import SpmBase



"""This module contains tests for the simulation devices in ophyd_devices"""

# pylint: disable: all
import os
import threading
import time
from unittest import mock

import h5py
import numpy as np
import pytest
from bec_lib import messages
from bec_lib.devicemanager import ScanInfo
from bec_lib.endpoints import MessageEndpoints
from bec_lib.file_utils import compile_file_components
from bec_server.device_server.tests.utils import DMMock
from ophyd import Device, Signal
from ophyd.status import wait as status_wait

from ophyd_devices.interfaces.protocols.bec_protocols import (
    BECDeviceProtocol,
    BECFlyerProtocol,
    BECPositionerProtocol,
    BECSignalProtocol,
)
from ophyd_devices.sim.sim_camera import SimCamera
from ophyd_devices.sim.sim_data import _safeint
from ophyd_devices.sim.sim_flyer import SimFlyer
from ophyd_devices.sim.sim_frameworks.h5_image_replay_proxy import H5ImageReplayProxy
from ophyd_devices.sim.sim_frameworks.slit_proxy import SlitProxy
from ophyd_devices.sim.sim_frameworks.stage_camera_proxy import StageCameraProxy
from ophyd_devices.sim.sim_monitor import SimMonitor, SimMonitorAsync
from ophyd_devices.sim.sim_positioner import SimLinearTrajectoryPositioner, SimPositioner
from ophyd_devices.sim.sim_signals import ReadOnlySignal
from ophyd_devices.sim.sim_test_devices import SimCameraWithPSIComponents
from ophyd_devices.sim.sim_utils import H5Writer, LinearTrajectory
from ophyd_devices.sim.sim_waveform import SimWaveform
from ophyd_devices.tests.utils import get_mock_scan_info
from ophyd_devices.utils.bec_device_base import BECDevice, BECDeviceBase

# pylint: disable=protected-access
# pylint: disable=no-member
# pylint: disable=too-many-arguments


@pytest.fixture(scope="function")
def waveform(name="waveform"):
    """Fixture for SimWaveform."""
    dm = DMMock()
    wave = SimWaveform(name=name, device_manager=dm)
    yield wave


@pytest.fixture(scope="function")
def signal(name="signal"):
    """Fixture for Signal."""
    sig = ReadOnlySignal(name=name, value=0)
    yield sig


@pytest.fixture(scope="function")
def monitor(name="monitor"):
    """Fixture for SimMonitor."""
    dm = DMMock()
    mon = SimMonitor(name=name, device_manager=dm)
    yield mon


@pytest.fixture(scope="function")
def camera(name="camera"):
    """Fixture for SimCamera."""
    dm = DMMock()
    cam = SimCamera(name=name, device_manager=dm, scan_info=ScanInfo)
    cam.filewriter = mock.MagicMock()
    cam.filewriter.compile_full_filename.return_value = ""
    yield cam


@pytest.fixture(scope="function")
def positioner(name="positioner"):
    """Fixture for SimPositioner."""
    dm = DMMock()
    pos = SimPositioner(name=name, device_manager=dm)
    yield pos


@pytest.fixture(scope="function")
def linear_traj_positioner(name="linear_traj_positioner"):
    """Fixture for SimLinearTrajectoryPositioner."""
    dm = DMMock()
    pos = SimLinearTrajectoryPositioner(name=name, device_manager=dm)
    yield pos


@pytest.fixture(scope="function")
def async_monitor(name="async_monitor"):
    """Fixture for SimMonitorAsync."""
    dm = DMMock()
    mon = SimMonitorAsync(name=name, device_manager=dm)
    yield mon


@pytest.fixture(scope="function")
def h5proxy_fixture(camera: SimCamera, name="h5proxy"):
    """Fixture for SimCamera."""
    dm = camera.device_manager
    proxy = H5ImageReplayProxy(name=name, device_manager=dm)
    yield proxy, camera


@pytest.fixture(scope="function")
def slitproxy_fixture(camera, name="slit_proxy"):
    """Fixture for SimCamera."""
    dm = camera.device_manager
    proxy = SlitProxy(name=name, device_manager=dm)
    samx = SimPositioner(name="samx", device_manager=dm)
    yield proxy, camera, samx


@pytest.fixture(scope="function")
def stage_camera_proxy_fixture(camera, name="stage_camera_proxy"):
    """Fixture for SimCamera."""
    dm = camera.device_manager
    proxy = StageCameraProxy(name=name, device_manager=dm)
    samx = SimPositioner(name="samx", limits=[-50, 50], device_manager=dm)
    samy = SimPositioner(name="samy", limits=[-50, 50], device_manager=dm)
    for device in (camera, proxy, samx, samy):
        device_mock = mock.MagicMock()
        device_mock.obj = device
        device_mock.enabled = True
        dm.devices[device.name] = device_mock
    proxy._update_device_config(
        {camera.name: {"signal_name": "image", "ref_motors": [samx.name, samy.name]}}
    )
    camera._registered_proxies.update({proxy.name: camera.image.name})
    proxy.enabled = True
    yield proxy, camera, samx, samy


@pytest.fixture(scope="function")
def flyer(name="flyer"):
    """Fixture for SimFlyer."""
    dm = DMMock()
    fly = SimFlyer(name=name, device_manager=dm)
    yield fly


def test_camera_with_sim_init():
    """Test to see if the sim init parameters are passed to the device"""
    dm = DMMock()
    sim = SimCamera(name="sim", device_manager=dm)
    assert sim.sim._model.value == "gaussian"
    model = "constant"
    params = {
        "amplitude": 300,
        "noise": "uniform",
        "noise_multiplier": 1,
        "hot_pixel_coords": [[0, 0], [50, 50]],
        "hot_pixel_types": ["fluctuating", "constant"],
        "hot_pixel_values": [2.0, 2.0],
    }
    sim = SimCamera(name="sim", device_manager=dm, sim_init={"model": model, "params": params})
    assert sim.sim._model.value == model
    assert sim.sim.params == params


def test_monitor_with_sim_init():
    """Test to see if the sim init parameters are passed to the device"""
    dm = DMMock()
    sim = SimMonitor(name="sim", device_manager=dm)
    assert sim.sim._model._name == "constant"
    model = "GaussianModel"
    params = {
        "amplitude": 500,
        "center": 5,
        "sigma": 4,
        "noise": "uniform",
        "noise_multiplier": 1,
        "ref_motor": "samy",
    }
    sim = SimMonitor(name="sim", device_manager=dm, sim_init={"model": model, "params": params})
    assert sim.sim._model._name == model.strip("Model").lower()
    diff_keys = set(sim.sim.params.keys()) - set(params.keys())
    for k in params:
        assert sim.sim.params[k] == params[k]


def test_signal__init__(signal):
    """Test the BECProtocol class"""
    assert isinstance(signal, BECSignalProtocol)


def test_monitor__init__(monitor):
    """Test the __init__ method of SimMonitor."""
    assert isinstance(monitor, SimMonitor)
    assert isinstance(monitor, BECSignalProtocol)


def test_camera__init__(camera):
    """Test the __init__ method of SimMonitor."""
    assert isinstance(camera, SimCamera)
    assert isinstance(camera, BECDeviceProtocol)


def test_positioner__init__(positioner):
    """Test the __init__ method of SimPositioner."""
    assert isinstance(positioner, SimPositioner)
    assert isinstance(positioner, BECPositionerProtocol)


def test_flyer__init__(flyer):
    """Test the __init__ method of SimFlyer."""
    assert isinstance(flyer, SimFlyer)
    assert isinstance(flyer, BECFlyerProtocol)


def test_init_async_monitor(async_monitor):
    """Test the __init__ method of SimMonitorAsync."""
    assert isinstance(async_monitor, SimMonitorAsync)
    assert isinstance(async_monitor, BECDeviceProtocol)


@pytest.mark.parametrize("center", [-10, 0, 10])
def test_monitor_readback(monitor, center):
    """Test the readback method of SimMonitor."""
    motor_pos = 0
    monitor.device_manager.add_device(name="samx", value=motor_pos)
    for model_name in monitor.sim.get_models():
        monitor.sim.select_model(model_name)
        monitor.sim.params["noise_multipler"] = 10
        monitor.sim.params["ref_motor"] = "samx"
        if "c" in monitor.sim.params:
            monitor.sim.params["c"] = center
        elif "center" in monitor.sim.params:
            monitor.sim.params["center"] = center
        assert isinstance(monitor.read()[monitor.name]["value"], monitor.BIT_DEPTH)
        expected_value = _safeint(monitor.sim._model.eval(monitor.sim._model_params, x=motor_pos))
        print(expected_value, monitor.read()[monitor.name]["value"])
        tolerance = (
            monitor.sim.params["noise_multipler"] + 1
        )  # due to ceiling in calculation, but maximum +1int
        assert np.isclose(
            monitor.read()[monitor.name]["value"],
            expected_value,
            atol=monitor.sim.params["noise_multipler"] + 1,
        )


@pytest.mark.parametrize("amplitude, noise_multiplier", [(0, 1), (100, 10), (1000, 50)])
def test_camera_readback(camera, amplitude, noise_multiplier):
    """Test the readback method of SimMonitor."""
    for model_name in camera.sim.get_models():
        camera.sim.select_model(model_name)
        camera.sim.params = {"noise_multiplier": noise_multiplier}
        camera.sim.params = {"amplitude": amplitude}
        camera.sim.params = {"noise": "poisson"}
        assert camera.image.get().shape == camera.SHAPE
        assert isinstance(camera.image.get()[0, 0], camera.BIT_DEPTH)
        camera.sim.params = {"noise": "uniform"}
        camera.sim.params = {"hot_pixel_coords": []}
        camera.sim.params = {"hot_pixel_values": []}
        camera.sim.params = {"hot_pixel_types": []}
        assert camera.image.get().shape == camera.SHAPE
        assert isinstance(camera.image.get()[0, 0], camera.BIT_DEPTH)
        assert (camera.image.get() <= (amplitude + noise_multiplier + 1)).all()


def test_positioner_move(positioner):
    """Test the move method of SimPositioner."""
    positioner.move(0).wait()
    assert np.isclose(
        positioner.read()[positioner.name]["value"], 0, atol=positioner.tolerance.get()
    )
    positioner.move(10).wait()
    assert np.isclose(
        positioner.read()[positioner.name]["value"], 10, atol=positioner.tolerance.get()
    )


@pytest.mark.timeout(30)
def test_positioner_motor_is_moving_signal(positioner):
    """Test that motor is moving is 0 and 1 while (not) moving"""

    recorded_data = []
    cid = None

    init_velocity = positioner.velocity.get()

    def motor_is_moving_cb(value: any, obj, **kwargs):
        data = obj.read()[f"{obj.name}_motor_is_moving"]["value"]
        recorded_data.append(data)

    try:
        cid = positioner.subscribe(motor_is_moving_cb, event_type="readback", run=False)

        status = positioner.move(-20)
        status.wait()

        positioner.velocity.set(10)

        status = positioner.move(20)
        status.wait()

        # Check that motor was moving and motor_is_moving switched to "1"
        assert any(recorded_data)

        # Check that motor is not moving and motor_is_moving switched back to "0"
        assert recorded_data[-1] == 0
    finally:
        # Restore initial velocity, remove subscription
        positioner.velocity.set(init_velocity)
        if cid is not None:
            positioner.unsubscribe(cid)


@pytest.mark.parametrize(
    "initial_position, final_position, max_velocity, acceleration",
    [(0, 100, 5, 20), (0, 1, 5, 20)],  # Trapezoidal profile  # Triangular profile
)
def test_linear_traj(initial_position, final_position, max_velocity, acceleration):
    """Test the LinearTrajectory class"""
    initial_time = time.time()
    trajectory = LinearTrajectory(
        initial_position, final_position, max_velocity, acceleration, initial_time
    )

    # Test acceleration phase
    t1 = initial_time + trajectory.time_accel / 2  # Halfway through acceleration phase
    pos1 = trajectory.position(t1)
    expected_pos1 = initial_position + 0.5 * acceleration * (trajectory.time_accel / 2) ** 2
    assert np.isclose(pos1, expected_pos1), f"Expected {expected_pos1}, got {pos1}"

    # Test constant velocity phase
    if trajectory.time_const_vel > 0:
        t2 = (
            initial_time + trajectory.time_accel + trajectory.time_const_vel / 2
        )  # Halfway through constant velocity phase
        pos2 = trajectory.position(t2)
        expected_pos2 = (
            initial_position
            + trajectory.distance_to_max_velocity
            + max_velocity * (t2 - initial_time - trajectory.time_accel)
        )
        assert np.isclose(pos2, expected_pos2), f"Expected {expected_pos2}, got {pos2}"

    # Test deceleration phase
    t3 = (
        initial_time + trajectory.total_time - trajectory.time_decel / 2
    )  # Halfway through deceleration phase
    pos3 = trajectory.position(t3)
    t_decel = t3 - (initial_time + trajectory.time_accel + trajectory.time_const_vel)
    expected_pos3 = final_position - 0.5 * acceleration * (trajectory.time_decel / 2) ** 2
    assert np.isclose(pos3, expected_pos3), f"Expected {expected_pos3}, got {pos3}"

    # Test end
    t4 = initial_time + trajectory.total_time + 0.1  # Slightly after end
    pos4 = trajectory.position(t4)
    assert pos4 == final_position
    assert trajectory.ended


def test_sim_linear_trajectory_positioner(linear_traj_positioner):
    vel = 5  # velocity 5 m.s^-1
    acc = 20  # acceleration 20 m.s^-2
    linear_traj_positioner.velocity.set(vel)
    linear_traj_positioner.acceleration.set(vel / acc)  # acctime 250 ms
    linear_traj_positioner.update_frequency = 100
    assert linear_traj_positioner.position == 0

    t0 = time.time()
    trajectory = LinearTrajectory(0, 50, vel, acc, t0)
    t2 = (
        t0 + trajectory.time_accel + trajectory.time_const_vel / 2
    )  # Halfway through constant velocity phase
    decel_distance = trajectory.position(t0 + trajectory.time_accel)
    expected_pos = trajectory.position(t2) + decel_distance

    linear_traj_positioner.move(50)
    # move is non-blocking, so sleep until it is time to stop:
    time.sleep(t2 - t0)
    linear_traj_positioner.stop()

    # ensure position is ok
    assert pytest.approx(linear_traj_positioner.position - expected_pos, abs=1e-1) == 0


@pytest.mark.parametrize("proxy_active", [True, False])
def test_sim_camera_proxies(camera, proxy_active):
    """Test mocking compute_method with framework class"""
    camera.device_manager.add_device("test_proxy")
    if proxy_active:
        camera._registered_proxies["test_proxy"] = camera.image.name
    else:
        camera._registered_proxies = {}
    proxy = camera.device_manager.devices["test_proxy"]
    mock_method = mock.MagicMock()
    mock_obj = proxy.obj
    mock_obj.lookup = mock.MagicMock()
    mock_obj.lookup.return_value = {camera.name: {"method": mock_method, "args": 1, "kwargs": 1}}
    camera.image.read()
    if proxy_active:
        assert len(mock_obj.lookup.mock_calls) > 0
    elif not proxy_active:
        assert len(mock_obj.lookup.mock_calls) == 0


def test_BECDeviceBase():
    # Test the BECDeviceBase class
    bec_device_base = BECDeviceBase(name="test")
    assert isinstance(bec_device_base, BECDevice)
    assert bec_device_base.connected is True
    signal = Signal(name="signal")
    assert isinstance(signal, BECDevice)
    device = Device(name="device")
    assert isinstance(device, BECDevice)


def test_h5proxy(h5proxy_fixture):
    """Test h5 camera proxy read from h5 file"""
    msg = messages.ScanStatusMessage(
        scan_id="test",
        num_points=10,
        scan_number=1,
        status="open",
        info={},
        scan_parameters={"frames_per_trigger": 1, "exp_time": 1},
    )
    h5proxy, camera = h5proxy_fixture
    mock_proxy = mock.MagicMock()
    camera.device_manager.devices.update({h5proxy.name: mock_proxy})
    mock_proxy.enabled = True
    mock_proxy.obj = h5proxy
    fname = os.path.expanduser("tests/test_data/h5_test_file.h5")
    h5entry = "entry/data/data"
    with h5py.File(fname, "r") as f:
        data = f[h5entry][...]
    # pylint: disable=protected-access
    h5proxy._update_device_config(
        {camera.name: {"signal_name": "image", "file_source": fname, "h5_entry": h5entry}}
    )
    camera._registered_proxies.update({h5proxy.name: camera.image.name})
    camera.sim.params = {"noise": "none", "noise_multiplier": 0}
    # pylint: disable=no-member
    camera.image_shape.set(data.shape[1:])
    with (
        mock.patch.object(camera.file_utils, "get_full_path", return_value="/tmp/path"),
        mock.patch.object(camera.scan_info, "msg", return_value=msg),
    ):
        camera.stage()
        img = camera.image.get()
        assert (img == data[0, ...]).all()
        camera.unstage()


def test_slitproxy(slitproxy_fixture):
    """Test slit proxy to compute readback from readback of positioner samx"""
    proxy, camera, samx = slitproxy_fixture
    for dev_name, dev in proxy.device_manager.devices.items():
        camera.device_manager.devices.update({dev_name: mock.MagicMock()})
        camera.device_manager.devices.get(dev_name).obj = dev
        camera.device_manager.devices.get(dev_name).enabled = True
    px_size = 0.5
    slitwidth = 2
    proxy._update_device_config(
        {
            camera.name: {
                "signal_name": "image",
                "center_offset": [0, 0],
                "covariance": [[1000, 500], [200, 1000]],
                "pixel_size": px_size,
                "ref_motors": [samx.name],
                "slit_width": [slitwidth],
                "motor_dir": [0],
            }
        }
    )
    camera._registered_proxies.update({proxy.name: camera.image.name})
    mock_proxy = mock.MagicMock()
    mock_samx = mock.MagicMock()
    mock_camera = mock.MagicMock()
    camera.device_manager.devices.update(
        {proxy.name: mock_proxy, samx.name: mock_samx, camera.name: mock_camera}
    )

    mock_proxy.enabled = True
    mock_samx.enabled = True
    mock_camera.enabled = True
    mock_camera.obj = camera
    mock_samx.obj = samx
    mock_proxy.obj = proxy
    camera.sim.params = {"noise": "none", "noise_multiplier": 0, "hot_pixel_values": [0, 0, 0]}
    samx.delay = 0
    samx_pos = 0
    samx.move(samx_pos)
    proxy._gaussian_blur_sigma = 0
    img = camera.image.get()
    edges = (
        int(img.shape[0] // 2 - samx_pos / px_size - slitwidth / (2 * px_size)),
        int(img.shape[0] // 2 + samx_pos / px_size + slitwidth / (2 * px_size)),
    )
    assert (img[:, : edges[0]] == 0).all()
    assert (img[:, edges[1] :] == 0).all()
    samx_pos = 13.3
    samx.move(samx_pos)
    img = camera.image.get()
    edges = (
        int(img.shape[0] // 2 + samx_pos / px_size - slitwidth / (2 * px_size)),
        int(img.shape[0] // 2 + samx_pos / px_size + slitwidth / (2 * px_size)),
    )
    assert (img[:, : edges[0]] == 0).all()
    assert (img[:, edges[1] :] == 0).all()


def test_proxy_config_and_props_stay_in_sync(h5proxy_fixture: tuple[H5ImageReplayProxy, SimCamera]):
    h5proxy, cam = h5proxy_fixture
    h5proxy._update_device_config(
        {
            cam.name: {
                "signal_name": "image",
                "file_source": "test first thing",
                "h5_entry": "/entry/data/data",
            }
        }
    )
    h5proxy.file_source = "test different thing"
    assert h5proxy.config[cam.name]["file_source"] == h5proxy.file_source
    h5proxy.h5_entry = "/entry/data/data_000001"
    assert h5proxy.config[cam.name]["h5_entry"] == h5proxy.h5_entry


def test_stage_camera_proxy_image_moves_with_samx_and_samy(
    stage_camera_proxy_fixture: tuple[StageCameraProxy, SimCamera, SimPositioner, SimPositioner],
):
    """Test camera stage proxy to compute readback from readback of positioner samx and samy"""
    proxy, camera, samx, samy = stage_camera_proxy_fixture

    proxy.stage()
    image_at_0: np.ndarray = camera.image.get()
    image_at_0_again: np.ndarray = camera.image.get()
    assert np.array_equal(image_at_0, image_at_0_again)
    samx.move(-10).wait()
    image_at_x_10 = camera.image.get()
    assert not np.array_equal(image_at_0, image_at_x_10)
    samy.move(-10).wait()
    image_at_x_10_y_10 = camera.image.get()
    assert not np.array_equal(image_at_x_10, image_at_x_10_y_10)


def test_stage_camera_proxy_image_shape(
    stage_camera_proxy_fixture: tuple[StageCameraProxy, SimCamera, SimPositioner, SimPositioner],
):
    """Make sure that the produced image has the same shape as the detector being proxied"""
    proxy, camera, samx, samy = stage_camera_proxy_fixture
    test_shape = (102, 77)
    camera.image_shape.set(test_shape).wait()
    proxy.stage()
    image: np.ndarray = camera.image.get()
    assert image.shape == (*reversed(test_shape), 3)


def test_cam_stage_h5writer(camera):
    """Test the H5Writer class"""
    file_dir = None
    suffix = None
    msg = messages.ScanStatusMessage(
        scan_id="test",
        num_points=10,
        scan_number=1,
        status="open",
        info={},
        scan_parameters={
            "frames_per_trigger": 1,
            "exp_time": 1,
            "system_config": {"file_directory": file_dir, "file_suffix": suffix},
            "file_components": compile_file_components(
                base_path="./test", scan_nr=1, file_directory=file_dir, user_suffix=suffix
            ),
        },
    )
    with (
        mock.patch.object(camera, "h5_writer") as mock_h5_writer,
        mock.patch.object(camera, "_run_subs") as mock_run_subs,
        mock.patch.object(camera.scan_info, "msg", return_value=msg),
        mock.patch.object(
            camera.file_utils, "get_full_path", return_value="./data/test_file_camera.h5"
        ),
    ):
        # camera.scan_info.msg.num_points = 10
        # camera.scan_info.msg.scan_parameters["frames_per_trigger"] = 1
        # camera.scan_info.msg.scan_parameters["exp_time"] = 1
        camera.stage()
        assert mock_h5_writer.on_stage.call_count == 0
        camera.unstage()
        camera.write_to_disk.put(True)
        camera.stage()
        calls = [mock.call(file_path="./data/test_file_camera.h5", h5_entry="/entry/data/data")]
        assert mock_h5_writer.on_stage.mock_calls == calls
        # mock_h5_writer.prepare


def test_cam_complete(camera):
    """Test the complete method of SimCamera."""
    finished_event = threading.Event()

    def finished_cb():
        finished_event.set()

    with mock.patch.object(camera, "h5_writer") as mock_h5_writer:
        status = camera.complete()
        status_wait(status)
        assert status.done is True
        assert status.success is True
        assert mock_h5_writer.on_complete.call_count == 0
        camera.write_to_disk.put(True)
        status = camera.complete()
        status.add_callback(finished_cb)
        finished_event.wait()
        assert mock_h5_writer.on_complete.call_count == 1


def test_cam_trigger(camera):
    """Test the trigger method of SimCamera."""
    with mock.patch.object(camera, "h5_writer") as mock_h5_writer:
        data = []
        status = camera.trigger()
        status_wait(status)
        assert status.done is True
        assert status.success is True
        assert mock_h5_writer.receive_data.call_count == 0
        camera.write_to_disk.put(True)
        status = camera.trigger()
        status_wait(status)
        assert mock_h5_writer.receive_data.call_count == 1
        status = camera.trigger()
        status_wait(status)
        assert mock_h5_writer.receive_data.call_count == 2


def test_h5writer(tmp_path):
    """Test the H5Writer class"""

    h5_writer = H5Writer()
    h5_writer.data_container = [np.array([0, 1, 2, 3, 4])]
    fp = tmp_path / "test.h5"
    h5_writer.on_stage(file_path=fp, h5_entry="entry/data/data")
    assert h5_writer.data_container == []
    assert h5_writer.file_path == fp
    assert h5_writer.h5_entry == "entry/data/data"

    data = np.array([0, 1])
    h5_writer.receive_data(data)
    assert h5_writer.data_container == [data]
    new_data = np.array([3, 4])
    h5_writer.receive_data(new_data)
    assert h5_writer.data_container == [data, new_data]
    h5_writer.receive_data(new_data)
    assert h5_writer.data_container == []
    h5_writer.receive_data(new_data)
    h5_writer.on_complete()
    assert h5_writer.data_container == []


def test_async_monitor_stage(async_monitor):
    """Test the stage method of SimMonitorAsync."""
    async_monitor.stage()
    assert async_monitor.data_buffer["value"] == []
    assert async_monitor.data_buffer["timestamp"] == []


def test_async_monitor_prep_random_interval(async_monitor):
    """Test the stage method of SimMonitorAsync."""
    async_monitor.prep_random_interval()
    assert async_monitor._counter == 0
    assert async_monitor.current_trigger.get() == 0
    assert 0 < async_monitor._random_send_interval < 10


def test_async_monitor_complete(async_monitor):
    """Test the on_complete method of SimMonitorAsync."""
    with (
        mock.patch.object(async_monitor, "_send_data_to_bec") as mock_send,
        mock.patch.object(async_monitor, "prep_random_interval") as mock_prep,
    ):
        status = async_monitor.complete()
        status_wait(status)
        assert status.done is True
        assert status.success is True
        assert mock_send.call_count == 0
        async_monitor.data_buffer["value"].append(0)
        status = async_monitor.complete()
        status_wait(status)
        assert status.done is True
        assert status.success is True
        assert mock_send.call_count == 1


def test_async_mon_on_trigger(async_monitor):
    """Test the on_trigger method of SimMonitorAsync."""
    with (mock.patch.object(async_monitor, "_send_data_to_bec") as mock_send,):
        async_monitor.on_stage()
        upper_limit = async_monitor._random_send_interval
        for ii in range(1, upper_limit + 1):
            status = async_monitor.on_trigger()
            status_wait(status)
            assert async_monitor.current_trigger.get() == ii
        assert mock_send.call_count == 1


def test_async_mon_send_data_to_bec(async_monitor):
    """Test the _send_data_to_bec method of SimMonitorAsync."""
    async_monitor.scan_info = get_mock_scan_info(device=async_monitor)
    async_monitor.data_buffer.update({"value": [0, 5], "timestamp": [0, 0]})
    with mock.patch.object(async_monitor.connector, "xadd") as mock_xadd:
        async_monitor._send_data_to_bec()
        dev_msg = messages.DeviceMessage(
            signals={async_monitor.readback.name: async_monitor.data_buffer},
            metadata={"async_update": {"type": "add", "max_shape": [None]}},
        )

        call = [
            mock.call(
                MessageEndpoints.device_async_readback(
                    scan_id=async_monitor.scan_info.msg.scan_id, device=async_monitor.name
                ),
                {"data": dev_msg},
                expire=async_monitor._stream_ttl,
            )
        ]
        assert mock_xadd.mock_calls == call
        assert async_monitor.data_buffer["value"] == []


def test_positioner_updated_timestamp(positioner):
    """Test the updated_timestamp method of SimPositioner."""
    positioner.sim.sim_state[positioner.name]["value"] = 1
    readback = positioner.read()[positioner.name]
    timestamp = readback["timestamp"]
    assert readback["value"] == 1
    positioner.sim.sim_state[positioner.name]["value"] = 5
    readback = positioner.read()[positioner.name]
    assert readback["value"] == 5
    assert readback["timestamp"] > timestamp


def test_waveform(waveform):
    """Test the SimWaveform class"""
    waveform.sim.select_model("GaussianModel")
    waveform.sim.params = {"amplitude": 500, "center": 500, "sigma": 10}
    data = waveform.waveform.get()
    assert isinstance(data, np.ndarray)
    assert data.shape == waveform.SHAPE
    assert np.isclose(np.argmax(data), 500, atol=5)
    waveform.waveform_shape.put(50)
    data = waveform.waveform.get()
    for model in waveform.sim.get_all_sim_models():
        waveform.sim.select_model(model)
        waveform.waveform.get()
    # Now also test the async readback
    mock_run_subs = waveform._run_subs = mock.MagicMock()
    waveform.scan_info = get_mock_scan_info(device=waveform)
    waveform.scan_info.msg.scan_id = "test"
    status = waveform.trigger()
    timer = 0
    while not status.done:
        time.sleep(0.1)
        timer += 0.1
        if timer > 5:
            raise TimeoutError("Trigger did not complete")
    assert status.done is True
    assert mock_run_subs.call_count == 1


@pytest.mark.parametrize(
    "mode, mock_data, expected_calls",
    [
        (
            "add",
            np.zeros(5),
            [{"sub_type": "device_monitor_1d", "value": np.zeros(5)}, {"value": np.zeros(5)}],
        )
    ],
)
def test_waveform_update_modes(waveform, mode, mock_data, expected_calls):
    """Test the add and add_slice update modes of the SimWaveform class"""
    waveform.sim.select_model("GaussianModel")
    waveform.sim.params = {"amplitude": 500, "center": 500, "sigma": 10}
    with pytest.raises(ValueError):
        waveform.async_update.put("invalid_mode")
    # Use add mode
    waveform.async_update.put(mode)
    with (
        mock.patch.object(waveform, "_run_subs") as mock_run_subs,
        mock.patch.object(waveform, "_send_async_update") as mock_send_async_update,
        mock.patch.object(waveform.waveform, "get", return_value=mock_data),
    ):

        status = waveform.trigger()
        status_wait(status, timeout=10)  # Raise if times out
        assert status.done is True
        # Run subs
        assert mock_run_subs.call_args[1]["sub_type"] == expected_calls[0]["sub_type"]
        assert np.array_equal(mock_run_subs.call_args[1]["value"], expected_calls[0]["value"])
        # Send async update
        assert np.array_equal(
            mock_send_async_update.call_args[1]["value"], expected_calls[1]["value"]
        )


@pytest.mark.parametrize(
    "mode, index, expected_md",
    [
        (
            "add_slice",
            0,
            {"async_update": {"type": "add_slice", "index": 0, "max_shape": [None, 100]}},
        ),
        ("add_slice", None, {"async_update": {"type": "add", "max_shape": [None, 200]}}),
        ("add", 0, {"async_update": {"type": "add", "max_shape": [None]}}),
    ],
)
def test_waveform_send_async_update(waveform, mode, index, expected_md):
    """Test the send_async_update method of SimWaveform."""
    max_shape = expected_md["async_update"]["max_shape"]
    if len(max_shape) > 1:
        wv_shape = max_shape[1]
    else:
        wv_shape = 100
    waveform.waveform_shape.put(wv_shape)
    waveform.async_update.put(mode)
    waveform.scan_info = get_mock_scan_info(device=waveform)
    value = np.random.rand(wv_shape)
    waveform._send_async_update(index=index, value=value)
    reading = waveform.data.read()
    assert (
        reading[waveform.data.name]["value"].metadata["async_update"] == expected_md["async_update"]
    )
    assert reading[waveform.data.name]["value"].signals[waveform.data.name]["value"].shape == (
        wv_shape,
    )


#####################################
### Test PSiComponent test device ###
#####################################


@pytest.fixture
def test_device():
    dev = SimCameraWithPSIComponents(name="test_device")
    yield dev


def test_simulation_sim_camera_with_psi_component(test_device):
    """Test the simulation test device with PSI components."""
    assert test_device.name == "test_device"
    assert all(
        element in test_device._signals
        for element in ["preview_2d", "preview_1d", "file_event", "progress", "dynamic_signal"]
    )
    # No signals are shown when read is called on the device
    assert test_device.read() == {}



"""Module for simulated monitor devices."""

import numpy as np
from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, Kind, StatusBase

from ophyd_devices.interfaces.base_classes.psi_device_base import PSIDeviceBase
from ophyd_devices.sim.sim_data import SimulatedDataMonitor
from ophyd_devices.sim.sim_signals import ReadOnlySignal, SetableSignal
from ophyd_devices.utils import bec_utils

logger = bec_logger.logger


class SimMonitor(ReadOnlySignal):
    """
    A simulated device mimic any 1D Axis (position, temperature, beam).

    It's readback is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataMonitor, more details on defaults
    within the simulation class.

    >>> monitor = SimMonitor(name="monitor")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent,
                              passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe().
                              Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation,
                              check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this
                              signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional.
                              Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation,
                              device_manager is passed on automatically.

    """

    USER_ACCESS = ["sim", "registered_proxies"]

    sim_cls = SimulatedDataMonitor
    BIT_DEPTH = np.uint32

    def __init__(
        self,
        name,
        *,
        precision: int = 3,
        sim_init: dict = None,
        parent=None,
        kind: Kind = None,
        device_manager=None,
        **kwargs,
    ):
        self.precision = precision
        self.sim_init = sim_init
        self.device_manager = device_manager
        self.sim = self.sim_cls(parent=self, **kwargs)
        self._registered_proxies = {}

        super().__init__(
            name=name,
            parent=parent,
            kind=kind,
            value=self.BIT_DEPTH(0),
            compute_readback=True,
            sim=self.sim,
            **kwargs,
        )
        if self.sim_init:
            self.sim.set_init(self.sim_init)

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies


class SimMonitorAsyncControl(Device):
    """SimMonitor Sync Control Device"""

    USER_ACCESS = ["sim", "registered_proxies", "async_update"]

    sim_cls = SimulatedDataMonitor
    BIT_DEPTH = np.uint32

    readback = Cpt(ReadOnlySignal, value=BIT_DEPTH(0), kind=Kind.hinted, compute_readback=True)
    current_trigger = Cpt(SetableSignal, value=BIT_DEPTH(0), kind=Kind.config)
    async_update = Cpt(SetableSignal, value="extend", kind=Kind.config)

    SUB_READBACK = "readback"
    SUB_PROGRESS = "progress"
    _default_sub = SUB_READBACK

    def __init__(self, name, *, sim_init: dict = None, parent=None, device_manager=None, **kwargs):
        if device_manager:
            self.device_manager = device_manager
        else:
            self.device_manager = bec_utils.DMMock()
        self.connector = self.device_manager.connector
        self.sim_init = sim_init
        self.sim = self.sim_cls(parent=self, **kwargs)
        self._registered_proxies = {}

        super().__init__(name=name, parent=parent, **kwargs)
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name
        self._data_buffer = {"value": [], "timestamp": []}
        if self.sim_init:
            self.sim.set_init(self.sim_init)

    @property
    def data_buffer(self) -> list:
        """Buffer for data to be sent asynchronously."""
        return self._data_buffer

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies


class SimMonitorAsync(PSIDeviceBase, SimMonitorAsyncControl):
    """
    A simulated device to mimic the behaviour of an asynchronous monitor.

    During a scan, this device will send data not in sync with the point ID to BEC,
    but buffer data and send it in random intervals.s
    """

    def __init__(
        self, name: str, scan_info=None, parent: Device = None, device_manager=None, **kwargs
    ) -> None:
        super().__init__(
            name=name, scan_info=scan_info, parent=parent, device_manager=device_manager, **kwargs
        )
        self._stream_ttl = 1800
        self._random_send_interval = None
        self._counter = 0
        self.prep_random_interval()

    def on_connected(self):
        self.current_trigger.subscribe(self._progress_update, run=False)

    def clear_buffer(self):
        """Clear the data buffer."""
        self.data_buffer["value"].clear()
        self.data_buffer["timestamp"].clear()

    def prep_random_interval(self):
        """Prepare counter and random interval to send data to BEC."""
        self._random_send_interval = np.random.randint(1, 10)
        self.current_trigger.set(0).wait()
        self._counter = self.current_trigger.get()

    def on_stage(self):
        """Prepare the device for staging."""
        self.clear_buffer()
        self.prep_random_interval()

    def on_complete(self) -> StatusBase:
        """Prepare the device for completion."""

        def complete_action():
            if self.data_buffer["value"]:
                self._send_data_to_bec()

        status = self.task_handler.submit_task(complete_action)
        return status

    def _send_data_to_bec(self) -> None:
        """Sends bundled data to BEC"""
        async_update = self.async_update.get()
        if async_update not in ["extend", "append"]:
            raise ValueError(f"Invalid async_update value for device {self.name}: {async_update}")

        metadata = None
        if async_update == "extend":
            metadata = {"async_update": {"type": "add", "max_shape": [None]}}
        elif async_update == "append":
            metadata = {"async_update": {"type": "add", "max_shape": [None, None]}}

        msg = messages.DeviceMessage(
            signals={self.readback.name: self.data_buffer}, metadata=metadata
        )
        self.connector.xadd(
            MessageEndpoints.device_async_readback(
                scan_id=self.scan_info.msg.scan_id, device=self.name
            ),
            {"data": msg},
            expire=self._stream_ttl,
        )
        self.clear_buffer()

    def on_trigger(self):
        """Prepare the device for triggering."""

        def trigger_action():
            """Trigger actions"""
            self.data_buffer["value"].append(self.readback.get())
            self.data_buffer["timestamp"].append(self.readback.timestamp)
            self._counter += 1
            self.current_trigger.set(self._counter).wait()
            if self._counter % self._random_send_interval == 0:
                self._send_data_to_bec()

        status = self.task_handler.submit_task(trigger_action)
        return status

    def _progress_update(self, value: int, **kwargs):
        """Update the progress of the device."""
        max_value = self.scan_info.msg.num_points
        # pylint: disable=protected-access
        self._run_subs(
            sub_type=self.SUB_PROGRESS,
            value=value,
            max_value=max_value,
            done=bool(max_value == value),
        )

    def on_stop(self):
        """Stop the device."""
        self.task_handler.shutdown()



import threading
import time
from unittest.mock import Mock

import pytest
from ophyd.status import DeviceStatus, StatusBase, StatusTimeoutError


def test_ophyd_status():
    device = Mock()
    device.name.return_value = "test"
    st = DeviceStatus(device)
    assert isinstance(st, StatusBase)

    cb = Mock()

    st = StatusBase(timeout=1)
    assert isinstance(st._callback_thread, threading.Thread)
    st.add_callback(cb)
    with pytest.raises(StatusTimeoutError):
        time.sleep(1.1)
        st.wait()
    cb.assert_called_once()
    cb.reset_mock()

    st = StatusBase()
    assert st._callback_thread is None
    st.add_callback(cb)
    st.set_finished()
    cb.assert_called_once()
    cb.reset_mock()
    st.wait()

    st = StatusBase(settle_time=1)
    st.add_callback(cb)
    assert st._callback_thread is None
    st.set_finished()
    assert cb.call_count == 0
    time.sleep(0.5)
    assert cb.call_count == 0  # not yet!
    time.sleep(0.6)
    cb.assert_called_once()
    cb.reset_mock()
    st.wait()

    class TestException(RuntimeError):
        pass

    st = StatusBase()
    st.add_callback(cb)
    st.set_exception(TestException())
    cb.assert_called_once()
    with pytest.raises(TestException):
        st.wait()



import threading
from unittest import mock

import ophyd
import pytest

from ophyd_devices.devices.delay_generator_645 import (
    DelayGenerator,
    DelayGeneratorError,
    TriggerSource,
)
from ophyd_devices.tests.utils import (
    MockPV,
    patch_dual_pvs,
    patch_functions_required_for_connection,
)


@pytest.fixture(scope="function")
def mock_ddg():
    name = "ddg"
    prefix = "X12SA-CPCL-DDG3:"
    with mock.patch.object(ophyd, "cl") as mock_cl:
        mock_cl.get_pv = MockPV
        mock_cl.thread_class = threading.Thread
        ddg = DelayGenerator(name=name, prefix=prefix)
        patch_functions_required_for_connection(ddg)
        patch_dual_pvs(ddg)
        yield ddg


def test_ddg_init(mock_ddg):
    """This test the initialization of the DelayGenerator"""
    assert mock_ddg.name == "ddg"
    assert mock_ddg.prefix == "X12SA-CPCL-DDG3:"


def test_set_trigger(mock_ddg):
    """This test the set_trigger method of the DelayGenerator"""
    mock_ddg.set_trigger(TriggerSource.SINGLE_SHOT)
    assert mock_ddg.source.get() == 5
    mock_ddg.set_trigger(TriggerSource.INTERNAL)
    assert mock_ddg.source.get() == 0


def test_burst_enable(mock_ddg):
    """This test the burst_enable method of the DelayGenerator"""
    count = 10
    delay = 0.1
    period = 0.2

    mock_ddg.burst_enable(count=count, delay=delay, period=period)
    assert mock_ddg.burstMode.get() == 1
    assert mock_ddg.burstCount.get() == count
    assert mock_ddg.burstDelay.get() == delay
    assert mock_ddg.burstPeriod.get() == period
    assert mock_ddg.burstConfig.get() == 0
    with pytest.raises(DelayGeneratorError):
        delay = -1
        mock_ddg.burst_enable(count=count, delay=delay, period=period)
    with pytest.raises(DelayGeneratorError):
        delay = 0
        period = 0
        mock_ddg.burst_enable(count=count, delay=delay, period=period)


def test_check_if_ddg_okay(mock_ddg):
    """This test the is_ddg_okay method of the DelayGenerator"""
    # Test for when the status is okay
    mock_ddg.status._read_pv.mock_data = "STATUS OK"
    assert mock_ddg.check_if_ddg_okay() is None
    # Test for when the status is not okay
    mock_ddg.status._read_pv.mock_data = "STATUS NOT OK"
    with pytest.raises(DelayGeneratorError):
        mock_ddg.check_if_ddg_okay()
